#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include <unistd.h>

#include <xlic/licensing.h>

#ifndef XLIC_VERSION
#define XLIC_VERSION "development version (built: " __DATE__ " " __TIME__ ")"
#endif

#ifndef XLICD_VALID_MULTIPLIER
#define XLICD_VALID_MULTIPLIER 86400 // 24 hours, in seconds
#endif

inline const char * to_string(xlic_license_state state) noexcept {
	switch (state) {
		case XLIC_STATE_INACTIVE: return "Inactive";
		case XLIC_STATE_ACTIVE: return "Active";
		case XLIC_STATE_GRACE: return "Grace";
		case XLIC_STATE_INVALID: return "Invalid";
		default: return "Unknown";
	}
}

inline const char * to_string(xlic_license_type type) noexcept {
	switch (type) {
		case XLIC_TYPE_PHYSICAL: return "Physical";
		case XLIC_TYPE_VIRTUAL: return "Virtual";
		default: return "Unknown";
	}
}

inline std::string to_string(time_t value) {
	char buffer[200]; // probably enough? (per manpage example)
	return std::string(buffer, strftime(buffer, sizeof(buffer), "%F %T", localtime(&value)));
}

inline int print_usage() noexcept {
	puts(
		"Usage: xlic [COMMAND] [ARGUMENT]\n"
		"Xorcom product licensing " XLIC_VERSION ".\n"
		"Copyright (C) 2017 Xorcom Ltd.\n"
		"\n"
		"Available COMMANDs:\n"
		"  fetch        Download licenses for this machine.\n"
		"  list         Display all licenses currently active.\n"
		"  register     Attempt to register license where ARGUMENT is a license key.\n"
		"  remove       Remove license where ARGUMENT is a license key.\n"
		"  clear        Remove all licenses from this system.\n"
		"\n"
	);
	return EXIT_FAILURE;
}

inline int print_human_error(const int code) noexcept {
	puts(xlic_errstr(code));
	return code;
}

inline void print_license(const xlic_license & license) noexcept {
	printf(
		"Name              : %s\n"
		"Description       : %s\n"
		"State             : %s\n"
		"Type              : %s\n"
		"License Key       : %s\n"
		"Valid After       : %s\n"
		"Valid Until       : %s\n"
		"Attributes        :\n",
		license.name,
		license.description,
		to_string(license.state),
		to_string(license.type),
		license.key,
		to_string(license.valid_after).c_str(),
		(license.valid_for > 0) ? to_string(license.valid_after + (license.valid_for * XLICD_VALID_MULTIPLIER)).c_str() : "Forever"
	);
	for (size_t i = 0; i < license.attribute_count; ++i) {
		printf(
			"- %-20s = %s\n",
			license.attributes[i].name,
			license.attributes[i].value
		);
	}
	puts("");
}

inline void print_licenses(const xlic_license_list * list) noexcept {
	if (list && list->count > 0) {
		printf(
			"Licenses          : %zu\n\n",
			list->count
		);
		for (size_t i = 0; i < list->count; ++i) {
			print_license(list->licenses[i]);
		}
	} else {
		puts("No licenses present on this system");
	}
}

int main(int argc, char ** argv) {
	const char * socket_filename = XLICD_SOCKETFILE;
	int opt = 0;
	while ((opt = ::getopt(argc, argv, "s:")) != -1) {
		switch (opt) {
			case 's': { socket_filename = optarg; break; }
			case '?': {
				printf("? %c\n", optopt);
				return print_usage();
			}
			default: { return EXIT_FAILURE; }
		}
	}
	argc -= optind;
	argv += optind;
	if (argc > 0) {
		if (strcmp(argv[0], "fetch") == 0) {
			const auto result = xlic_fetch(socket_filename);
			if (result == XLIC_RESULT_SUCCESS) {
				puts("Fetch completed successfully");
				return EXIT_SUCCESS;
			}
			return print_human_error(result);
		} else if (strcmp(argv[0], "list") == 0) {
			xlic_license_list * list = nullptr;
			const auto result = xlic_licenses(socket_filename, &list);
			if (result == XLIC_RESULT_SUCCESS) {
				puts("List completed successfully");
				print_licenses(list);
				return EXIT_SUCCESS;
			}
			return print_human_error(result);
		} else if (strcmp(argv[0], "register") == 0) {
			if (argc > 1) {
				const auto result =	xlic_register_license(socket_filename, argv[1]);
				if (result == XLIC_RESULT_SUCCESS) {
					puts("Registration successful");
					return EXIT_SUCCESS;
				}
				return print_human_error(result);
			}
			//NOTE: no license key provided
		} else if (strcmp(argv[0], "remove") == 0) {
			if (argc > 1) {
				const auto result = xlic_unregister_license(socket_filename, argv[1]);
				if (result == XLIC_RESULT_SUCCESS) {
					puts("Removal successful");
					return EXIT_SUCCESS;
				}
				return print_human_error(result);
			}
			//NOTE: no license key provided
		} else if (strcmp(argv[0], "clear") == 0) {
			const auto result = xlic_clear(socket_filename);
			if (result == XLIC_RESULT_SUCCESS) {
				puts("Clear completed successfully");
				return EXIT_SUCCESS;
			}
			return print_human_error(result);
		}
	}
	return print_usage();
}
