#include <xlic/licensing.h>
#include <php.h>

#define XLIC_MODULE_NAME "CompletePBX"
#define XLIC_MINIMUM_EXTENSIONS 3
#define XLIC_FAXMASTER_DEFAULT false
#define XLIC_EPM_PAIRING_DEFAULT false
#define XLIC_MINIMUM_ROOMS 5

#define nullptr NULL

inline const char * to_string(xlic_license_state state) {
	switch (state) {
		case XLIC_STATE_INACTIVE: { return "inactive"; }
		case XLIC_STATE_ACTIVE: { return "active"; }
		case XLIC_STATE_GRACE: { return "grace"; }
		case XLIC_STATE_INVALID: { return "invalid"; }
		default: { return "unknown"; }
	}
}

inline const char * to_string(xlic_license_type type) {
	switch (type) {
		case XLIC_TYPE_PHYSICAL: { return "physical"; }
		case XLIC_TYPE_VIRTUAL: { return "virtual"; }
		default: { return "unknown"; }
	}
}

#if PHP_VERSION_ID >= 50400 && PHP_VERSION_ID < 70000

typedef zval * array_type;
typedef zval * object_type;

inline void create_array(array_type & array) {
	array = nullptr;
	MAKE_STD_ZVAL(array);
	array_init(array);
}

inline void create_object(object_type & object) {
	object = nullptr;
	MAKE_STD_ZVAL(object);
	object_init(object);
}

inline void object_set(object_type & object, const char * property, const char * value) {
	add_property_string_ex(object, property, strlen(property) + 1, value, true);
}

inline void object_set(object_type & object, const char * property, xlic_license_state value) {
	add_property_string_ex(object, property, strlen(property) + 1, to_string(value), true);
}

inline void object_set(object_type & object, const char * property, xlic_license_type value) {
	add_property_string_ex(object, property, strlen(property) + 1, to_string(value), true);
}

inline void object_set(object_type & object, const char * property, time_t value) {
	add_property_long_ex(object, property, strlen(property) + 1, value);
}

inline void object_set(object_type & object, const char * property, size_t value) {
	add_property_long_ex(object, property, strlen(property) + 1, value);
}

inline void object_set(object_type & object, const char * property, array_type & array) {
	add_property_zval_ex(object, property, strlen(property) + 1, array);
}

inline void array_set(array_type & array, const char * key, const char * value) {
	add_assoc_string_ex(array, key, strlen(key) + 1, const_cast<char *>(value), true); // braindead PHP devs
}

inline void array_append(array_type & array, object_type & object) {
	add_next_index_zval(array, object);
}

#define RETURN_ARRAY(array) RETURN_ZVAL(array, true, false);

#elif PHP_VERSION_ID >= 70000

typedef zval array_type;
typedef zval object_type;

inline void create_array(array_type & array) {
	array_init(&array);
}

inline void create_object(object_type & object) {
	object_init(&object);
}

inline void object_set(object_type & object, const char * property, const char * value) {
	add_property_string_ex(&object, property, strlen(property), value);
}

inline void object_set(object_type & object, const char * property, xlic_license_state value) {
	add_property_string_ex(&object, property, strlen(property), to_string(value));
}

inline void object_set(object_type & object, const char * property, xlic_license_type value) {
	add_property_string_ex(&object, property, strlen(property), to_string(value));
}

inline void object_set(object_type & object, const char * property, time_t value) {
	add_property_long_ex(&object, property, strlen(property), value);
}

inline void object_set(object_type & object, const char * property, size_t value) {
	add_property_long_ex(&object, property, strlen(property), value);
}

inline void object_set(object_type & object, const char * property, array_type & array) {
	add_property_zval_ex(&object, property, strlen(property), &array);
}

inline void array_set(array_type & array, const char * key, const char * value) {
	add_assoc_string_ex(&array, key, strlen(key), const_cast<char *>(value)); // braindead PHP devs
}

inline void array_append(array_type & array, object_type & object) {
	add_next_index_zval(&array, &object);
}

#define RETURN_ARRAY(array) RETURN_ZVAL(&array, false, false);

#endif

int allowed_extensions() {
	xlic_license_list * list = nullptr;
	xlic_licenses(XLICD_SOCKETFILE, &list);
	if (list) {
		for (size_t i = 0; i < list->count; ++i) {
			const auto & license = list->licenses[i];
			if (license.state == XLIC_STATE_ACTIVE || license.state == XLIC_STATE_GRACE) {
				if (strcmp(license.name, "cpbx-nolimit-extensions") == 0) {
					return INT_MAX; // 32-bit signed max for PHP safety
				} else if (strcmp(license.name, "completepbx") == 0) {
					for (size_t j = 0; j < license.attribute_count; ++j) {
						const auto & attribute = license.attributes[j];
						if (strcmp(attribute.name, "extensions") == 0) {
							int len = 0;
							size_t value = 0;
							if (sscanf(attribute.value, "%zu%n", &value, &len) == 1 && len) {
								return value;
							}
						}
					}
				} else {
					int len = 0;
					size_t value = 0;
					if (sscanf(license.name, "cpbx-%zu-extensions%n", &value, &len) == 1 && len) {
						return value;
					}
				}
			}
		}
		xlic_free(list);
	}
	return XLIC_MINIMUM_EXTENSIONS;
}

int allowed_rooms() {
	xlic_license_list * list = nullptr;
	xlic_licenses(XLICD_SOCKETFILE, &list);
	if (list) {
		for (size_t i = 0; i < list->count; ++i) {
			const auto & license = list->licenses[i];
			if (license.state == XLIC_STATE_ACTIVE || license.state == XLIC_STATE_GRACE) {
				if (strcmp(license.name, "complete-concierge") == 0) {
					for (size_t j = 0; j < license.attribute_count; ++j) {
						const auto & attribute = license.attributes[j];
						if (strcmp(attribute.name, "rooms") == 0) {
							return atoi(attribute.value);
						}
					}
				}
			}
		}
		xlic_free(list);
	}
	return XLIC_MINIMUM_ROOMS;
}

bool list_contains(const char * name, bool _default = false) {
	xlic_license_list * list = nullptr;
	xlic_licenses(XLICD_SOCKETFILE, &list);
	if (list) {
		for (size_t i = 0; i < list->count; ++i) {
			const auto & license = list->licenses[i];
			if (license.state == XLIC_STATE_ACTIVE || license.state == XLIC_STATE_GRACE) {
				if (strcmp(license.name, name) == 0) {
					return true;
				}
			}
		}
		xlic_free(list);
	}
	return _default;
}

PHP_FUNCTION(ombu_license_list) {
	array_type licenses;
	create_array(licenses);
	xlic_license_list * list = nullptr;
	xlic_licenses(XLICD_SOCKETFILE, &list);
	if (list) {
		for (size_t i = 0; i < list->count; ++i) {
			object_type object;
			create_object(object);
			const auto & license = list->licenses[i];
			object_set(object, "state", license.state);
			object_set(object, "type", license.type);
			object_set(object, "key", license.key);
			object_set(object, "name", license.name);
			object_set(object, "description", license.description);
			object_set(object, "valid_after", license.valid_after);
			object_set(object, "valid_for", license.valid_for);
			object_set(object, "grace_days", license.grace_days);
			array_type attributes;
			create_array(attributes);
			for (size_t j = 0; j < license.attribute_count; ++j) {
				const auto & attribute = license.attributes[j];
				array_set(attributes, attribute.name, attribute.value);
			}
			object_set(object, "attributes", attributes);
			array_append(licenses, object);
		}
		xlic_free(list);
	}
	RETURN_ARRAY(licenses);
}

PHP_FUNCTION(ombu_fetch_licenses) {
	xlic_fetch(XLICD_SOCKETFILE);
}

PHP_FUNCTION(ombu_allowed_extensions) {
	bool faxmaster = XLIC_FAXMASTER_DEFAULT;
	RETURN_LONG(allowed_extensions());
}

PHP_FUNCTION (ombu_faxmaster) {
	RETURN_BOOL(list_contains("cpbx-faxmaster", XLIC_FAXMASTER_DEFAULT) || allowed_extensions() > XLIC_MINIMUM_EXTENSIONS);
}

PHP_FUNCTION (ombu_epm_pairing) {
	RETURN_BOOL(list_contains("cpbx-epm-pairing", XLIC_EPM_PAIRING_DEFAULT));
}

PHP_FUNCTION (ombu_update_licenses) {
	xlic_update(XLICD_SOCKETFILE);
}

PHP_FUNCTION (cc_allowed_rooms) {
	RETURN_LONG(allowed_rooms());
}

zend_function_entry xlic_functions[] = {
	PHP_FE(ombu_license_list, nullptr)
	PHP_FE(ombu_fetch_licenses, nullptr)
	PHP_FE(ombu_allowed_extensions, nullptr)
	PHP_FE(ombu_faxmaster, nullptr)
	PHP_FE(ombu_update_licenses, nullptr)
	PHP_FE(ombu_epm_pairing, nullptr)
	PHP_FE(cc_allowed_rooms, nullptr)
	PHP_FE_END
};

zend_module_entry xlic_module_entry = {
	STANDARD_MODULE_HEADER,
	XLIC_MODULE_NAME,
	xlic_functions,
	nullptr,
	nullptr,
	nullptr,
	nullptr,
	nullptr,
	XLIC_VERSION,
	STANDARD_MODULE_PROPERTIES
};

extern "C" ZEND_DLEXPORT zend_module_entry * get_module() {
	return &xlic_module_entry;
}
