#ifndef __XLIC_TYPES__
#define __XLIC_TYPES__

#include <xlic/licensing.h>
#include <xtl/os.hpp>
#include <xtl/hash_method.hpp>
#include <xtl/istream.hpp>
#include <xtl/json.hpp>
#include <xtl/ostream.hpp>
#include <xtl/rsa.hpp>

#include <cstdint>
#include <cstring>
#include <set>
#include <string>
#include <map>
#include <vector>

#ifndef XLIC_SERVER_HOSTNAME
#define XLIC_SERVER_HOSTNAME "license.xorcom.com"
#endif

#ifndef XLIC_USERAGENT
#define XLIC_USERAGENT "xlic/2.0"
#endif

#ifndef XLIC_CONTENT_TYPE
#define XLIC_CONTENT_TYPE "application/xlic"
#endif

namespace xlic {

class invalid_public_key : public xtl::exception {
public:
	inline invalid_public_key(const char * file, int line)
	: exception(file, line, "%s", "Invalid public key") {}
};

class invalid_signature : public xtl::exception {
public:
	inline invalid_signature(const char * file, int line)
	: exception(file, line, "%s", "Invalid signature") {}
};

class invalid_license_key : public xtl::exception {
public:
	inline invalid_license_key(const char * file, int line)
	: exception(file, line, "%s", "Invalid license key") {}
};

class license_key {
public:
	inline void * value() noexcept { return m_value; }
	inline const void * value() const noexcept { return m_value; }
	inline size_t size() const noexcept { return sizeof(m_value); }

	bool operator<(const xlic::license_key & other) const noexcept;

private:
	uint8_t m_value[20];
};

struct serial_number : public std::string {

	inline serial_number() {}
	inline explicit serial_number(const std::string & string)
	: std::string(string) {}

	inline serial_number & operator=(const std::string & string) {
		assign(string);
		return *this;
	}

	inline size_t size() const { return length() + 1; }
};

enum class license_type : uint8_t {
	unknown = 0,
	virtual_license = XLIC_TYPE_VIRTUAL,
	physical_license = XLIC_TYPE_PHYSICAL,
};

enum class license_state : uint8_t {
	unknown = XLIC_STATE_UNKNOWN,
	inactive = XLIC_STATE_INACTIVE,
	active = XLIC_STATE_ACTIVE,
	grace = XLIC_STATE_GRACE,
	invalid = XLIC_STATE_INVALID,
	force_invalid,
};

class attribute_list : public std::map<std::string, std::string> {
public:
	inline size_t count() const { return map::size(); }
	size_t size() const;
};

class license {
public:
	xlic::license_key license_key;
	xtl::rsa_signature signature;
	xlic::license_state state;
	xlic::license_type type;
	time_t valid_after;
	time_t registered_on;
	time_t updated_on;
	time_t grace_start;
	uint32_t valid_for;
	uint32_t grace_days;
	uint64_t nonce;
	std::string name;
	std::string description;
	xlic::attribute_list attributes;

	inline license() noexcept
	: state(xlic::license_state::unknown)
	, type(xlic::license_type::unknown)
	, valid_after(0)
	, registered_on(0)
	, updated_on(0)
	, grace_start(0)
	, valid_for(0)
	, grace_days(0)
	, nonce(0)
	{}

	size_t size() const;

	bool verify(const xtl::rsa_public_key & public_key, const xlic::serial_number & serial_number) const;

	bool operator<(const xlic::license & other) const noexcept;
};

class license_list : public std::map<xlic::license_key, xlic::license> {
public:
	inline size_t count() const { return map::size(); }
	size_t size() const;
};

class mac_address_list : public std::set<xtl::mac_address> {
public:
	inline size_t count() const { return set::size(); }
	size_t size() const;
};

enum class protocol : uint8_t {
	unknown = 0,
	version_1 = 1,
	latest = version_1,
};

enum class packet_type : uint8_t {
	unknown = 0,
	fetch_licenses = 1,
	register_license = 2,
	unregister_license = 3,
	list_licenses = 4,
	license = 5,
	license_list = 6,
	success = 7,
	error = 8,
	update_licenses = 9,
	clear_licenses = 10,
};

enum class result_code : int {
	unknown = XLIC_RESULT_UNKNOWN,
	success = XLIC_RESULT_SUCCESS,
	generic_error = XLIC_RESULT_GENERIC_ERROR,
	server_error = XLIC_RESULT_SERVER_ERROR,
	daemon_error = XLIC_RESULT_DAEMON_ERROR,
	rejected = XLIC_RESULT_REJECTED,
	invalidkey = XLIC_RESULT_INVALIDKEY,
	nolicense =XLIC_RESULT_NOLICENSE,
	keyexists = XLIC_RESULT_KEYEXISTS,
	foreign = XLIC_RESULT_FOREIGN,
};

typedef uint32_t request_id;
typedef uint32_t packet_size;
typedef int32_t extension_count;

enum { packet_header_size = sizeof(xlic::packet_size) + sizeof(xlic::packet_type) + sizeof(xlic::request_id) };

// from string

void from_string(const std::string & string, xlic::license_type & license_type);
void from_string(const std::string & string, xlic::license_state & license_state);

// from <type>

template<typename T> T from_base64(const std::string & string);
template<typename T> T from_json(const xtl::json & js);

// to string

std::string to_string(xtl::rsa_public_key & public_key);
std::string to_string(const xlic::license_key & license_key);
std::string to_string(const xlic::license_type & license_type);
std::string to_string(const xlic::license_state & license_state);
std::string to_string(const xtl::rsa_signature & signature);
std::string to_string(const time_t & timestamp);
std::string to_string(const uint32_t & number);
std::string to_string(const uint64_t & number);

// Stream operators

xtl::hash_method & operator+=(xtl::hash_method & hash_method, const xlic::mac_address_list & mac_address_list);

inline xtl::hash_method & operator+=(xtl::hash_method & hash_method, const std::string & string) {
	return hash_method.append(string.data(), string.length());
}

inline xtl::verify_method & operator+=(xtl::verify_method & verify_method, const std::string & string) {
	return verify_method.append(string.data(), string.length());
}

inline xtl::ostream & operator<<(xtl::ostream & stream, const std::string & string) {
	return xtl::write(stream, string.data(), string.length());
}

xtl::istream & operator>>(xtl::istream & stream, xlic::serial_number & serial_number);
xtl::ostream & operator<<(xtl::ostream & stream, const xlic::serial_number & serial_number);
xtl::istream & operator>>(xtl::istream & stream, xlic::attribute_list & attribute_list);
xtl::ostream & operator<<(xtl::ostream & stream, const xlic::attribute_list & attribute_list);
xtl::istream & operator>>(xtl::istream & stream, xlic::license & license);
xtl::ostream & operator<<(xtl::ostream & stream, const xlic::license & license);
xtl::istream & operator>>(xtl::istream & stream, xlic::license_list & license_list);
xtl::ostream & operator<<(xtl::ostream & stream, const xlic::license_list & license_list);
xtl::istream & operator>>(xtl::istream & stream, xlic::mac_address_list & mac_address_list);
xtl::ostream & operator<<(xtl::ostream & stream, const xlic::mac_address_list & mac_address_list);

inline xtl::istream & operator>>(xtl::istream & stream, xtl::rsa_signature & signature) {
	return xtl::read(stream, signature.value(), signature.size());
}

inline xtl::ostream & operator<<(xtl::ostream & stream, const xtl::rsa_signature & signature) {
	return xtl::write(stream, signature.value(), signature.size());
}

inline xtl::istream & operator>>(xtl::istream & stream, xtl::rsa_public_key & public_key) {
	return xtl::read(stream, public_key.value(), public_key.size());
}

inline xtl::ostream & operator<<(xtl::ostream & stream, xtl::rsa_public_key & public_key) {
	return xtl::write(stream, public_key.value(), public_key.size());
}

inline xtl::istream & operator>>(xtl::istream & stream, xlic::license_key & license_key) {
	return xtl::read(stream, license_key.value(), license_key.size());
}

inline xtl::ostream & operator<<(xtl::ostream & stream, const xlic::license_key & license_key) {
	return xtl::write(stream, license_key.value(), license_key.size());
}

inline xtl::istream & operator>>(xtl::istream & stream, xlic::license_type & license_type) {
	return xtl::read(stream, &license_type, sizeof(license_type));
}

inline xtl::ostream & operator<<(xtl::ostream & stream, xlic::license_type license_type) {
	return xtl::write(stream, &license_type, sizeof(license_type));
}

inline xtl::istream & operator>>(xtl::istream & stream, xlic::license_state & license_state) {
	return xtl::read(stream, &license_state, sizeof(license_state));
}

inline xtl::ostream & operator<<(xtl::ostream & stream, xlic::license_state license_state) {
	return xtl::write(stream, &license_state, sizeof(license_state));
}

inline xtl::istream & operator>>(xtl::istream & stream, xlic::protocol & protocol) {
	return xtl::read(stream, &protocol, sizeof(protocol));
}

inline xtl::ostream & operator<<(xtl::ostream & stream, xlic::protocol protocol) {
	return xtl::write(stream, &protocol, sizeof(protocol));
}

inline xtl::istream & operator>>(xtl::istream & stream, xlic::packet_type & packet_type) {
	return xtl::read(stream, &packet_type, sizeof(packet_type));
}

inline xtl::ostream & operator<<(xtl::ostream & stream, xlic::packet_type packet_type) {
	return xtl::write(stream, &packet_type, sizeof(packet_type));
}

inline xtl::istream & operator>>(xtl::istream & stream, xlic::result_code & result_code) {
	return xtl::read(stream, &result_code, sizeof(result_code));
}

inline xtl::ostream & operator<<(xtl::ostream & stream, xlic::result_code result_code) {
	return xtl::write(stream, &result_code, sizeof(result_code));
}

inline xtl::istream & operator>>(xtl::istream & stream, xtl::mac_address & mac_address) {
	return xtl::read(stream, mac_address.value(), mac_address.size());
}

inline xtl::ostream & operator<<(xtl::ostream & stream, const xtl::mac_address & mac_address) {
	return xtl::write(stream, mac_address.value(), mac_address.size());
}

} // namespace xlic

#endif
