#include <xlic/client.hpp>

#include <xtl/exception.hpp>
#include <xtl/thread.hpp>

#if defined(DEBUG_CLIENT) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) printf("client[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) printf("client[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xlic {

client::client(const char * filename)
: m_socket(filename)
, m_next_request_id(0)
{
	fdebug("Created new client %p", (void *) this);
	m_socket.set_timeout(5000);
	m_socket << xlic::protocol::latest;
}

client::~client() noexcept {
	fdebug("Destroying client %p", (void *) this);
}

struct response_header {
	inline response_header(xtl::istream & stream)
	{
		stream >> response_size;
		if (response_size < xlic::packet_header_size) {
			throw xlic::client_exception(__FILE__, __LINE__, "Invalid response size");
		}
		stream >> response_type;
		stream >> response_id;
	}

	xlic::packet_size response_size;
	xlic::packet_type response_type;
	xlic::request_id response_id;
};

void client::clear_licenses() {
	debug("clear_licenses()");
	const auto request_id = m_next_request_id++;
	m_socket << xlic::packet_size(xlic::packet_header_size);
	m_socket << xlic::packet_type::clear_licenses;
	m_socket << request_id;
	const xlic::response_header header(m_socket);
	if (header.response_id != request_id) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response id");
	}
	if (header.response_type == xlic::packet_type::error) {
		xlic::result_code result_code;
		m_socket >> result_code;
		throw result_code;
	}
	if (header.response_type != xlic::packet_type::success) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response type");
	}
}

void client::fetch_licenses() {
	debug("fetch_licenses()");
	const auto request_id = m_next_request_id++;
	m_socket << xlic::packet_size(xlic::packet_header_size);
	m_socket << xlic::packet_type::fetch_licenses;
	m_socket << request_id;
	const xlic::response_header header(m_socket);
	if (header.response_id != request_id) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response id");
	}
	if (header.response_type == xlic::packet_type::error) {
		xlic::result_code result_code;
		m_socket >> result_code;
		throw result_code;
	}
	if (header.response_type != xlic::packet_type::success) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response type");
	}
}

void client::register_license(const std::string & key) {
	fdebug("register_license(%s)", key.c_str());
	const auto license_key = from_base64<xlic::license_key>(key);
	const auto request_id = m_next_request_id++;
	m_socket << xlic::packet_size(xlic::packet_header_size + license_key.size());
	m_socket << xlic::packet_type::register_license;
	m_socket << request_id;
	m_socket << license_key;
	const xlic::response_header header(m_socket);
	if (header.response_id != request_id) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response id");
	}
	if (header.response_type == xlic::packet_type::error) {
		xlic::result_code result_code;
		m_socket >> result_code;
		throw result_code;
	}
	if (header.response_type != xlic::packet_type::success) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response type");
	}
}

void client::unregister_license(const std::string & key) {
	fdebug("unregister_license(%s)", key.c_str());
	const auto license_key = from_base64<xlic::license_key>(key);
	const auto request_id = m_next_request_id++;
	m_socket << xlic::packet_size(xlic::packet_header_size + license_key.size());
	m_socket << xlic::packet_type::unregister_license;
	m_socket << request_id;
	m_socket << license_key;
	const xlic::response_header header(m_socket);
	if (header.response_id != request_id) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response id");
	}
	if (header.response_type == xlic::packet_type::error) {
		xlic::result_code result_code;
		m_socket >> result_code;
		throw result_code;
	}
	if (header.response_type != xlic::packet_type::success) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response type");
	}
}

license_list client::list_licenses() {
	debug("list_licenses()");
	const auto request_id = m_next_request_id++;
	m_socket << xlic::packet_size(xlic::packet_header_size);
	m_socket << xlic::packet_type::list_licenses;
	m_socket << request_id;
	const xlic::response_header header(m_socket);
	if (header.response_id != request_id) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response id");
	}
	if (header.response_type == xlic::packet_type::error) {
		xlic::result_code result_code;
		m_socket >> result_code;
		throw result_code;
	}
	if (header.response_type != xlic::packet_type::license_list) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response type");
	}
	xlic::license_list licenses;
	m_socket >> licenses;
	return licenses;
}

void client::update_licenses() {
	debug("update()");
	const auto request_id = m_next_request_id++;
	m_socket << xlic::packet_size(xlic::packet_header_size);
	m_socket << xlic::packet_type::update_licenses;
	m_socket << request_id;
	const xlic::response_header header(m_socket);
	if (header.response_id != request_id) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response id");
	}
	if (header.response_type == xlic::packet_type::error) {
		xlic::result_code result_code;
		m_socket >> result_code;
		throw result_code;
	}
	if (header.response_type != xlic::packet_type::success) {
		throw xlic::client_exception(__FILE__, __LINE__, "Invalid response type");
	}
}

} //namespace
