#include <cstdint>
#include <string>
#include <exception>
#include <memory>
#include <system_error>
#include <pthread.h>
#include <mysql/mysql.h>
#include <xlic/types.hpp>

#if defined(DEBUG_MYSQL) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) printf("mysql[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) printf("mysql[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

class mysql_exception : public std::exception {
public:
	inline mysql_exception(const char * reason, const unsigned int code = 0) noexcept
	: m_reason(reason)
	, m_code(code)
	{}

	inline const char * what() const noexcept override {
		return m_reason;
	}

	inline unsigned int code() const noexcept {
		return m_code;
	}

private:
	const char * m_reason;
	const unsigned int m_code;
};

inline mysql_exception make_exception(MYSQL * session) {
	return mysql_exception(mysql_error(session), mysql_errno(session));
}

inline mysql_exception make_exception(MYSQL_STMT * statement) {
	return mysql_exception(mysql_stmt_error(statement), mysql_stmt_errno(statement));
}

inline MYSQL_STMT * make_statement(MYSQL * session) {
	auto stmt = mysql_stmt_init(session);
	if (!stmt) {
		throw make_exception(session);
	}
	return stmt;
}

class mysql_statement {
public:
	inline mysql_statement(
		MYSQL * session,
		const std::string & query
	)
	: m_statement(make_statement(session), mysql_stmt_close)
	{
		if (mysql_stmt_prepare(m_statement.get(), query.c_str(), query.length()) != 0) {
			throw make_exception(m_statement.get());
		}
	}

	inline int execute() const {
		if (mysql_stmt_execute(m_statement.get()) != 0) {
			throw make_exception(m_statement.get());
		}
		int result;
		unsigned long real_length;
		MYSQL_BIND bind;
		bind.buffer_type = MYSQL_TYPE_LONG;
		bind.buffer = &result;
		bind.buffer_length = sizeof(result);
		bind.length = &real_length;
		bind.is_null = nullptr;
		bind.is_unsigned = false;
		bind.error = nullptr;
		if (mysql_stmt_bind_result(m_statement.get(), &bind) != 0) {
			throw make_exception(m_statement.get());
		}
		if (mysql_stmt_fetch(m_statement.get()) != 0) {
			throw make_exception(m_statement.get());
		}
		return result;
	}

private:
	std::unique_ptr<MYSQL_STMT, my_bool (*)(MYSQL_STMT *)> m_statement;
};

static pthread_mutex_t g_mysql_session_guard = PTHREAD_MUTEX_INITIALIZER;

inline MYSQL * make_session() {
	if (pthread_mutex_lock(&g_mysql_session_guard) != 0) {
		throw std::system_error(errno, std::generic_category());
	}
	auto session = mysql_init(nullptr);
	pthread_mutex_unlock(&g_mysql_session_guard);
	if (!session) {
		throw mysql_exception("Failed to initialize library");
	}
	return session;
}

class mysql_session {
private:
	std::unique_ptr<MYSQL, void (*)(MYSQL *)> m_session;

public:
	inline mysql_session(
		const std::string & database,
		const std::string & username,
		const std::string & password
	)
	: m_session(make_session(), mysql_close)
	{
		if (!mysql_real_connect(
			m_session.get(), nullptr, username.c_str(), password.c_str(),
			database.c_str(), 0, nullptr, 0
		)) {
			throw make_exception(m_session.get());
		}
	}

	inline mysql_statement prepare(const std::string & query) const {
		return mysql_statement(m_session.get(), query);
	}
};

namespace xlic {

extension_count count_extensions() noexcept {
	try {
		const auto session = mysql_session("ombutel", "ombutel", "ombutel");
		const auto stmt = session.prepare("select count(1) from ombu_extensions");
		return stmt.execute();
	} catch (const std::exception & e) {
		fdebug("%s", e.what());
		return -1;
	}
}

} // namespace xlic
