#include <xlic/server.hpp>

#include <xtl/cryptostream.hpp>
#include <xtl/exception.hpp>
#include <xtl/fs.hpp>
#include <xtl/filestream.hpp>
#include <xtl/lock_guard.hpp>
#include <xtl/os.hpp>
#include <xtl/sha256.hpp>
#include <xtl/interval_timer.hpp>
#include <xtl/https_stream.hpp>
#include <xlic/worker.hpp>
#include <xlic/json.hpp>
#include <xlic/rockey.hpp>
#include <zlib.h>

#ifndef XLIC_SERIAL_NUMBER
#include <auth-dongle/api.h>
#endif

#if defined(DEBUG_SERVER) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) printf("server[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) printf("server[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

#ifndef XLIC_LICENSE_FILENAME
#define XLIC_LICENSE_FILENAME "/etc/cpbx/license.bin"
#endif

#ifndef XLIC_SERIAL_FILENAME
#define XLIC_SERIAL_FILENAME "/etc/cpbx/serial"
#endif

#ifndef XLICD_CLOCK_SKEW_LIMIT
#define XLICD_CLOCK_SKEW_LIMIT 86400 // 24 hours, in seconds
#endif

#ifndef XLICD_UPDATE_LIMIT
#define XLICD_UPDATE_LIMIT 172800 // 48 hours, in seconds
#endif

#ifndef XLICD_GRACE_MULTIPLIER
#define XLICD_GRACE_MULTIPLIER 86400 // 24 hours, in seconds
#endif

#ifndef XLICD_VALID_MULTIPLIER
#define XLICD_VALID_MULTIPLIER 86400 // 24 hours, in seconds
#endif

#ifndef XLICD_UPDATE_INTERVAL
#define XLICD_UPDATE_INTERVAL 43200 // 12 hours, in seconds
#endif

#define HEADER_MAGIC 0x43494c58

namespace xlic {

inline void meditate(const char * psalm) {
	printf("Guru meditation %s\n", psalm);
}

struct v2_header {
	inline v2_header(xtl::istream & stream) {
		stream >> count;
		stream >> public_key;
	}

	uint32_t count;
	xtl::rsa_public_key public_key;
};

struct v2_license {
	inline v2_license(xtl::istream & stream) {
		stream >> last_update;
		stream >> expires;
		stream >> data_offset;
		stream >> name_length;
		stream >> description_length;
		stream >> license_key;
		stream >> signature;
	}

	uint64_t last_update;
	uint32_t expires;
	uint32_t data_offset;
	uint16_t name_length;
	uint16_t description_length;
	xlic::license_key license_key;
	xtl::rsa_signature signature;
};

struct file_header {
	inline file_header(xtl::istream & stream) {
		stream >> magic;
		stream >> version;
	}

	inline constexpr file_header(uint32_t magic, uint32_t version) noexcept
	: magic(magic), version(version) {}

	uint32_t magic;
	uint32_t version;
};

inline xtl::ostream & operator<<(xtl::ostream & stream, const xlic::file_header & header) {
	stream << header.magic;
	stream << header.version;
	return stream;
}

server::~server() noexcept = default;

inline void load_v2_license_file(
	xtl::istream & stream,
	xtl::rsa_public_key & public_key,
	xlic::license_list & licenses,
	time_t & last_update
) {
	struct auxillary {
		uint16_t name_length;
		uint16_t description_length;
		xlic::license * license;
	};
	const v2_header header(stream);
	fdebug("Loading v2 license file with %u license(s)", header.count);
	std::vector<auxillary> aux;
	aux.reserve(header.count);
	for (auto i = header.count;i > 0; --i) {
		const v2_license old_license(stream);
		const time_t license_last_update = old_license.last_update / 1000000;
		if (last_update < license_last_update) {
			last_update = license_last_update;
		}
		xlic::license license;
		license.license_key = old_license.license_key;
		license.signature = old_license.signature;
		license.valid_after = ::time(nullptr);
		license.valid_for = (old_license.expires > 0) ? old_license.expires - license.valid_after : 0;
		license.grace_days = 30;
		aux.push_back({
			old_license.name_length,
			old_license.description_length,
			&licenses.emplace(
				std::piecewise_construct,
				std::forward_as_tuple(license.license_key),
				std::forward_as_tuple(license)
			).first->second
		});
	}
	for (size_t i = 0; i < header.count; ++i) {
		const size_t name_len = aux[i].name_length;
		auto & name = aux[i].license->name;
		name.resize(name_len);
		xtl::read(stream, &name[0], name_len);
		const size_t description_length = aux[i].description_length;
		auto & description = aux[i].license->description;
		description.resize(description_length);
		xtl::read(stream, &description[0], description_length);
	}
	public_key = header.public_key;
}

inline std::vector<uint8_t> decode_decryption_key() {
	const size_t key_bytes = 64; // Taken from cryptostream.cpp
	const uint8_t encoded_key[key_bytes] = { // Cannot be static or compiler will generate relocation entry
		0xd3, 0x9e, 0xff, 0xc5, 0xca, 0xc9, 0xb9, 0xc0, 0xce, 0xca, 0x86, 0xd9, 0x0c, 0xae, 0xa3, 0x92,
		0x85, 0xb6, 0x95, 0x93, 0xf2, 0xb3, 0x09, 0xae, 0x9b, 0xf9, 0xf9, 0xec, 0x39, 0xdd, 0x40, 0xd9,
		0xb6, 0xc6, 0x19, 0xfa, 0x94, 0xa3, 0xbd, 0x3b, 0x7b, 0x74, 0xf6, 0x82, 0x07, 0x1b, 0xad, 0xc7,
		0xc4, 0xa4, 0x8b, 0x25, 0xde, 0x47, 0xd0, 0xd8, 0xb9, 0x51, 0xed, 0x2b, 0x6c, 0x89, 0x6e, 0x44,
	};
	// xor key with itself in reverse
	std::vector<uint8_t> decoded(sizeof(encoded_key));
	for (size_t i = 0; i < key_bytes; ++i) {
		decoded[i] = encoded_key[i] ^ encoded_key[key_bytes - i - 1];
	}
	return decoded;
}

inline bool match(const xlic::mac_address_list & left, const xlic::mac_address_list & right) {
	for (const auto & l : left) {
		for (const auto & r : right) {
			if (l == r) {
				return true;
			}
		}
	}
	return false;
}

inline void load_v3_license_file(
	bool have_dongle,
	xtl::iostream & file,
	xtl::rsa_public_key & public_key,
	const xlic::serial_number & serial_number,
	xlic::serial_number & foreign_serial_number,
	const xlic::mac_address_list & mac_addresses,
	xlic::mac_address_list & foreign_mac_addresses,
	xlic::license_list & licenses,
	time_t & last_update,
	bool & foreign_license_file
) {
	foreign_license_file = false;
	const size_t iv_size = xtl::crypto::iv_size();
	std::vector<uint8_t> iv(iv_size);
	xtl::read(file, iv.data(), iv_size);
	xtl::cryptostream stream(file);
	stream.set_read_iv(iv.data(), iv_size);
	stream.set_read_key(decode_decryption_key().data(), xtl::crypto::key_size());
	time_t last_save;
	xtl::sha256_digest stored_digest;
	stream >> last_save;
	stream >> last_update;
	stream >> public_key;
	stream >> foreign_serial_number;
	stream >> licenses;
	stream >> foreign_mac_addresses;
	stream >> stored_digest;
	fdebug("Loaded v3 license file with %zu license(s)", licenses.count());
	// Ignore foreign license file / manipulation when no licenses present
	if (licenses.count() == 0) {
		return;
	}
	// Try to detect clock manipulation
	const time_t now = ::time(nullptr);
	const time_t skew_limit = XLICD_CLOCK_SKEW_LIMIT;
	if (last_save > time_t(now + skew_limit)) {
		throw xlic::server_exception(__FILE__, __LINE__, "Clock skew detected");
	}
	// Try to detect tampering
	xtl::sha256 hash;
	hash += last_save;
	hash += last_update;
	hash += foreign_serial_number;
	hash += foreign_mac_addresses;
	const auto computed_digest = hash.digest();
	if (computed_digest != stored_digest) {
		throw xlic::server_exception(__FILE__, __LINE__, "License file validation failed");
	}
	if (have_dongle) {
		// Try to detect copying of license file
		if (serial_number != foreign_serial_number) {
			meditate("X0016");
			foreign_license_file = true;
			debug("Serial number mismatch");
		}
	}
	// Try to detect copying of license file
	if (!match(mac_addresses, foreign_mac_addresses)) {
		meditate("X0017");
		foreign_license_file = true;
		debug("MAC address mismatch");
	}
}

#ifdef XLIC_SERIAL_NUMBER

inline xlic::serial_number get_serial_number(bool & have_dongle) {
	have_dongle = true;
	return xlic::serial_number(XLIC_SERIAL_NUMBER);
}

#else // XLIC_SERIAL_NUMBER

inline bool get_serial_number_from_xorcom_dongle(xlic::serial_number & serial_number) {
	struct auth_dongle * auth_dongle = auth_dongle_setup(nullptr, nullptr, 5, 2000); // 5 retries, 2000ms timeout
	if (auth_dongle) {
		const char * serial = auth_dongle_read_serial(auth_dongle); // to be treated as strdup, returns null on error
		if (serial) {
			// Dongles apparently contain 64 zeroes before burning
			if (strlen(serial) != 64) {
				serial_number = serial;
			}
			free(const_cast<char *>(serial));
			auth_dongle_cleanup(auth_dongle);
			return true;
		} else {
			debug("Unable to access dongle (do we have one?)");
		}
		auth_dongle_cleanup(auth_dongle);
	} else {
		debug("Unable to access dongle (do we have one?)");
	}
	return false;
}

#ifdef USE_ROCKEY

inline bool get_serial_number_from_rockey_dongle(xlic::serial_number & serial_number) {
	try {
		const uint16_t p1 = 0x553c, p2 = 0xe190; // These need to be adjusted accordingly
		auto dongles = rockey::enumerate(p1, p2);
		if (dongles.size() == 0) return false; // No dongles found
		for (auto & dongle : dongles) {
			dongle.open(p1, p2);
			const auto info = dongle.read();
			const auto valid_dongle = dongle.is_xorcom_dongle(info.serial, info.magic);
			dongle.close();
			if (valid_dongle) {
				serial_number = info.serial;
				return true;
			}
		}
	} catch (const std::exception & e) {
		debug(e.what());
	}
	return false;
}

#endif // USE_ROCKEY

inline xlic::serial_number get_serial_number(bool & have_dongle) {
	xlic::serial_number serial_number;
	have_dongle = get_serial_number_from_xorcom_dongle(serial_number);
#ifdef USE_ROCKEY
	if (!have_dongle) {
		have_dongle = get_serial_number_from_rockey_dongle(serial_number);
	}
#endif // USE_ROCKEY
	return serial_number;
}

#endif // XLIC_SERIAL_NUMBER

inline xlic::mac_address_list get_mac_addresses() {
	xlic::mac_address_list mac_addresses;
	for (const xtl::network_interface & interface : xtl::os::interfaces()) {
		mac_addresses.emplace(interface.mac_address);
	}
	return mac_addresses;
}

inline void update_serial_file(const xlic::serial_number & serial_number) {
	xtl::filestream file(XLIC_SERIAL_FILENAME, xtl::filemode::writeonly);
	file.truncate();
	file << std::string(serial_number);
}

server::server(const std::string & filename)
: m_have_dongle(false)
, m_foreign_license_file(false)
, m_last_update(0)
, m_serial_number(get_serial_number(m_have_dongle))
, m_mac_addresses(get_mac_addresses())
, m_running(load_licenses())
, m_socket(filename)
, m_updater([this] () {
	xtl::interval_timer timer(XLICD_UPDATE_INTERVAL * 1000);
	while (m_running) {
		bool update_succeeded = false;
		try {
			if (!m_foreign_license_file) {
				update_licenses();
				update_succeeded = true;
			}
		} catch (const std::exception & e) {
			fprintf(stderr, "%s\n", e.what());
		}
		if (!update_succeeded) {
			// Update failed, update license file to allow clock skew detection
			xtl::lock_guard<xtl::mutex> guard(m_mutex);
			save_licenses();
		}
		timer.wait();
	}
})
{
	auto worker = std::make_unique<xlic::worker>(*this);
	m_workers.emplace(worker.get(), std::move(worker));
}

extension_count count_extensions() noexcept;

void server::update_licenses() {
	if (!m_foreign_license_file) {
		xtl::lock_guard<xtl::mutex> guard(m_mutex);
		const auto licenses_copy = m_licenses;
		const auto mac_addresses_copy = m_mac_addresses;
		guard.release();
		const auto serial_number_copy = serial_number();
		const auto extensions = count_extensions();
		xtl::https_stream stream(XLIC_SERVER_HOSTNAME);
		stream.post("/update");
		stream.useragent(XLIC_USERAGENT);
		stream.content_type(XLIC_CONTENT_TYPE);
		stream << xlic::json_encode(serial_number_copy, mac_addresses_copy, licenses_copy, extensions);
		stream.flush();
		const xtl::json js(stream);
		const auto serial_number = xlic::serial_number(js["serial_number"].string());
		const auto public_key = from_base64<xtl::rsa_public_key>(js["public_key"].string());
		const auto licenses = from_json<xlic::license_list>(js["licenses"]);
		update(serial_number, public_key, licenses);
	}
}

xlic::license_list server::licenses() {
	xtl::lock_guard<xtl::mutex> guard(m_mutex);
	xlic::license_list copy = m_licenses;
	// Clear foreign flag when there are no licenses
	if (m_licenses.count() == 0) {
		m_foreign_license_file = false;
	}
	const time_t now = ::time(nullptr);
	for (auto & pair : copy) {
		auto & license = pair.second;
		if (m_foreign_license_file) {
			license.state = xlic::license_state::grace;
			meditate("X0018");
			fdebug("License \"%s\" is foreign", license.name.c_str());
		} else if (license.state == xlic::license_state::force_invalid) {
			// License rejected by server on update
			license.state = xlic::license_state::grace;
			meditate("X0000");
			fdebug("License \"%s\" rejected by server on update", license.name.c_str());
		} else if (!license.verify(m_public_key, m_serial_number)) {
			// Signature does not compute
			license.state = xlic::license_state::grace;
			meditate("X0001");
			fdebug("Signature for \"%s\" does not compute", license.name.c_str());
		} else if (license.valid_after > now) {
			// License not yet active
			license.state = xlic::license_state::inactive;
			meditate("X0002");
			fdebug("License \"%s\" not yet active", license.name.c_str());
		} else if (license.valid_for != 0 && time_t(license.valid_after + (license.valid_for * XLICD_VALID_MULTIPLIER)) < now) {
			// License has expired
			license.state = xlic::license_state::grace;
			meditate("X0003");
			fdebug("License \"%s\" has expired", license.name.c_str());
			if (license.grace_days == 0) {
				license.state = xlic::license_state::invalid;
				meditate("X0004");
				debug("But has no grace period");
			} else if (time_t(license.valid_after + (license.valid_for * XLICD_VALID_MULTIPLIER) + (license.grace_days * XLICD_GRACE_MULTIPLIER)) < now) {
				// Grace period also expired
				license.state = xlic::license_state::invalid;
				meditate("X0005");
				debug("Grace period also expired");
			}
		} else if (license.type == xlic::license_type::virtual_license && time_t(m_last_update + XLICD_UPDATE_LIMIT) < now) {
			// (Virtual) Unable to contact server for extended period
			license.state = xlic::license_state::grace;
			meditate("X0006");
			debug("Unable to contact server for extended period");
			if (license.grace_days == 0) {
				license.state = xlic::license_state::invalid;
				meditate("X0007");
				fdebug("License \"%s\" has no grace period", license.name.c_str());
			} else if (time_t(m_last_update + (license.grace_days * XLICD_GRACE_MULTIPLIER)) < now) {
				// Could not contact server for entire grace period
				license.state = xlic::license_state::invalid;
				meditate("X0008");
				fdebug("Could not contact server for entire grace period of \"%s\"", license.name.c_str());
			}
		} else if (license.type == xlic::license_type::physical_license && !m_have_dongle) {
			// (Physical) No dongle detected on startup
			license.state = xlic::license_state::grace;
			meditate("X0009");
			debug("No dongle detected on startup");
		} else {
			license.state = xlic::license_state::active;
			meditate("X0010");
			fdebug("License \"%s\" is active", license.name.c_str());
		}
		if (license.state == xlic::license_state::grace) {
			if (license.grace_start == 0) {
				// Start grace period
				if (license.grace_days > 0) {
					m_licenses[license.license_key].grace_start = now;
					meditate("X0011");
					fdebug("Starting grace for \"%s\"", license.name.c_str());
				} else {
					license.state = xlic::license_state::invalid;
					meditate("X0012");
					fdebug("License \"%s\" has no grace period", license.name.c_str());
				}
			} else if (time_t(license.grace_start + (license.grace_days * XLICD_GRACE_MULTIPLIER)) < now) {
				// Grace period expired
				license.state = xlic::license_state::invalid;
				meditate("X0013");
				fdebug("Grace period expired for \"%s\"", license.name.c_str());
			}
		}
	}
	return copy;
}

xtl::rsa_signature server::signature(const xlic::license_key & license_key) const {
	xtl::lock_guard<xtl::mutex> guard(m_mutex);
	const auto it = m_licenses.find(license_key);
	if (it != m_licenses.end()) {
		return it->second.signature;
	}
	return xtl::rsa_signature();
}

void server::stop() {
	debug("Stopping server");
	m_running = false;
	for (const auto & pair : m_workers) {
		pair.second->stop();
	}
}

void server::add(
	const xlic::serial_number & serial_number,
	const xtl::rsa_public_key & public_key,
	const xlic::license & license
) {
	if (!m_foreign_license_file) {
		xtl::lock_guard<xtl::mutex> guard(m_mutex);
		m_serial_number = serial_number;
		m_public_key = public_key;
		m_licenses.emplace(license.license_key, license);
		m_last_update = ::time(nullptr);
		save_licenses();
	}
}

bool server::exists(const xlic::license_key & license_key) {
	xtl::lock_guard<xtl::mutex> guard(m_mutex);
	return (m_licenses.find(license_key) != m_licenses.end());
}

void server::remove_license(const xlic::license_key & license_key) {
	xtl::lock_guard<xtl::mutex> guard(m_mutex);
	m_licenses.erase(license_key);
	if (m_licenses.count() == 0) {
		// No more licenses, thus no more foreign licenses
		m_foreign_license_file = false;
	}
	save_licenses();
}

bool server::load_licenses() {
	try {
		xtl::filestream file(XLIC_LICENSE_FILENAME);
		fdebug("Loading licenses from %s", XLIC_LICENSE_FILENAME);
		const file_header header(file);
		if (header.magic != HEADER_MAGIC) {
			throw xlic::server_exception(__FILE__, __LINE__, "Not a license file");
		}
		switch (header.version) {
			case 2: {
				load_v2_license_file(file, m_public_key, m_licenses, m_last_update);
				break;
			}
			case 3: {
				load_v3_license_file(
					m_have_dongle,
					file,
					m_public_key,
					m_serial_number,
					m_foreign_serial_number,
					m_mac_addresses,
					m_foreign_mac_addresses,
					m_licenses,
					m_last_update,
					m_foreign_license_file
				);
				if (!m_have_dongle) {
					m_serial_number = m_foreign_serial_number;
				}
				break;
			}
			default: throw xlic::server_exception(__FILE__, __LINE__, "Invalid license file version");
		}
		update_serial_file(m_serial_number);
		return true;
	} catch (const xtl::file_not_found &) {
		fdebug("%s", "License file not found, continuing\n");
		return true;
	} catch (const xtl::end_of_stream &) {
		fdebug("%s", "License file truncated or corrupt, continuing\n");
		return true;
	} catch (const std::exception & e) {
		fprintf(stderr, "%s\n", e.what());
	}
	return false;
}

void server::save_licenses() {
	fdebug("Saving v3 license file with %zu licenses(s) to %s", m_licenses.count(), XLIC_LICENSE_FILENAME);
	const std::string tmp_filename(XLIC_LICENSE_FILENAME ".tmp");
	try {
		const size_t iv_size = xtl::crypto::iv_size();
		std::vector<uint8_t> iv(iv_size);
		xtl::crypto::randomize(iv.data(), iv_size);
		xtl::filestream file(tmp_filename, xtl::filemode::writeonly);
		file.truncate();
		file << file_header(HEADER_MAGIC, 3);
		xtl::write(file, iv.data(), iv_size);
		xtl::cryptostream stream(file);
		xtl::sha256 hash;
		stream.set_write_iv(iv.data(), iv_size);
		stream.set_write_key(decode_decryption_key().data(), xtl::crypto::key_size());
		const time_t last_save = ::time(nullptr);
		stream << last_save;
		stream << m_last_update;
		stream << m_public_key;
		if (m_foreign_license_file) {
			stream << m_foreign_serial_number;
		} else {
			stream << m_serial_number;
		}
		stream << m_licenses;
		if (m_foreign_license_file) {
			stream << m_foreign_mac_addresses;
		} else {
			stream << m_mac_addresses;
		}
		hash += last_save;
		hash += m_last_update;
		if (m_foreign_license_file) {
			hash += m_foreign_serial_number;
			hash += m_foreign_mac_addresses;
		} else {
			hash += m_serial_number;
			hash += m_mac_addresses;
		}
		stream << hash.digest();
		stream.flush();
		xtl::fs::rename(tmp_filename, XLIC_LICENSE_FILENAME);
		update_serial_file(m_serial_number);
	} catch (const std::exception & e) {
		fprintf(stderr, "%s\n", e.what());
	}
	xtl::fs::remove(tmp_filename);
}

void server::update(
	const xlic::serial_number & serial_number,
	const xtl::rsa_public_key & public_key,
	const xlic::license_list & licenses,
	bool replace
) {
	if (!m_foreign_license_file) {
		xtl::lock_guard<xtl::mutex> guard(m_mutex);
		m_serial_number = serial_number;
		m_public_key = public_key;
		if (replace) {
			// Replace only physical licenses by first removing and then re-inserting
			for (auto it = m_licenses.begin(); it != m_licenses.end();) {
				if (it->second.type == xlic::license_type::physical_license) {
					it = m_licenses.erase(it);
				} else {
					++it;
				}
			}
			for (auto & license : licenses) {
				m_licenses[license.second.license_key] = license.second;
			}
		} else {
			for (auto & license : m_licenses) {
				license.second.state = xlic::license_state::force_invalid;
			}
			for (const auto & license : licenses) {
				const auto it = m_licenses.find(license.second.license_key);
				if (it != m_licenses.end()) {
					m_licenses[license.second.license_key] = license.second;
				}
			}
		}
		m_last_update = ::time(nullptr);
		save_licenses();
	}
}

xlic::serial_number server::serial_number() {
	xtl::lock_guard<xtl::mutex> guard(m_mutex);
	bool have_dongle = false;
	const auto serial_number = get_serial_number(have_dongle);
	if (m_have_dongle || have_dongle) {
		if (m_serial_number.empty()) {
			meditate("X0014");
			debug("No dongle detected previously, updating serial number");
			m_serial_number = serial_number;
		} else if (m_serial_number != serial_number) {
			meditate("X0015");
			debug("Serial number has been changed");
			m_serial_number = serial_number;
		}
	}
	return m_serial_number;
}

} // namespace xlic
