#include <xlic/session.hpp>

#include <xtl/exception.hpp>
#include <xlic/clear_licenses.hpp>
#include <xlic/fetch_licenses.hpp>
#include <xlic/register_license.hpp>
#include <xlic/update_licenses.hpp>
#include <xlic/unregister_license.hpp>
#include <xlic/list_licenses.hpp>
#include <xlic/server.hpp>
#include <xlic/worker.hpp>

#if defined(DEBUG_SESSION) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) printf("session[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) printf("session[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xlic {

struct request_header {
	inline request_header(xtl::istream & stream)
	: payload_size(0)
	{
		stream >> request_size;
		stream >> request_type;
		stream >> request_id;

		if (request_size < xlic::packet_header_size) {
			throw xlic::session_exception(__FILE__, __LINE__, "Invalid request size");
		}
		payload_size = request_size - xlic::packet_header_size;
	}

	xlic::packet_size request_size;
	xlic::packet_type request_type;
	xlic::request_id request_id;
	size_t payload_size;
};

session::session(xlic::worker & worker, xtl::unix_socket && socket)
: m_worker(worker)
, m_socket(std::move(socket))
, m_reader([this] () {
	try {
		read_input();
	} catch (const std::exception & e) {
		fdebug("%s\n", e.what());
	}
	m_worker.close(this);
}) {
	fdebug("Created new session %p", (void *) this);
	m_socket.set_timeout(5000);
}

void session::read_input() {
	xlic::protocol protocol;
	m_socket >> protocol;
	if (protocol != xlic::protocol::latest) {
		throw xlic::session_exception(__FILE__, __LINE__, "Invalid protocol version");
	}
	for (;;) {
		const xlic::request_header header(m_socket);
		switch(header.request_type) {
			case xlic::packet_type::fetch_licenses: {
				m_worker.add(make_task<xlic::fetch_licenses>(this, m_worker, header.request_id));
				break;
			}
			case xlic::packet_type::register_license: {
				xlic::license_key license_key;
				if(header.payload_size < license_key.size()) {
					throw xlic::session_exception(__FILE__, __LINE__, "Invalid payload size");
				}
				m_socket >> license_key;
				m_worker.add(make_task<xlic::register_license>(this, m_worker, header.request_id, license_key));
				break;
			}
			case xlic::packet_type::unregister_license: {
				xlic::license_key license_key;
				if(header.payload_size < license_key.size()) {
					throw xlic::session_exception(__FILE__, __LINE__, "Invalid payload size");
				}
				m_socket >> license_key;
				m_worker.add(make_task<xlic::unregister_license>(this, m_worker, header.request_id, license_key));
				break;
			}
			case xlic::packet_type::list_licenses: {
				m_worker.add(make_task<xlic::list_licenses>(this, m_worker, header.request_id));
				break;
			}
			case xlic::packet_type::update_licenses: {
				m_worker.add(make_task<xlic::update_licenses>(this, m_worker, header.request_id));
				break;
			}
			case xlic::packet_type::clear_licenses: {
				m_worker.add(make_task<xlic::clear_licenses>(this, m_worker, header.request_id));
				break;
			}
			default: {
				throw xlic::session_exception(__FILE__, __LINE__, "Invalid request type");
			}
		}
	}
}

session::~session() noexcept {
	fdebug("Destroying session %p", (void *) this);
}

} // namespace xlic
