#include <xlic/types.hpp>

#include <xtl/base64.hpp>
#include <xtl/exception.hpp>
#include <algorithm>
#include <inttypes.h>

namespace xlic {

typedef uint8_t string_length;
typedef uint32_t array_count;

// strings

xtl::istream & read(xtl::istream & stream, std::string & string) {
	xlic::string_length length;
	stream >> length;
	string.resize(length);
	return xtl::read(stream, &string[0], length);
}

xtl::ostream & write(xtl::ostream & stream, const std::string & string) {
	const xlic::string_length length = std::min(string.length(), size_t(255));
	stream << length;
	return xtl::write(stream, string.data(), length);
}

// public key

template<> xtl::rsa_public_key from_base64<xtl::rsa_public_key>(const std::string & string) {
	xtl::rsa_public_key public_key;
	if(xtl::base64::decode(
		public_key.value(), public_key.size(),
		string.c_str(), string.length()
	) != public_key.size()) {
		throw xlic::invalid_public_key(__FILE__, __LINE__);
	}
	return public_key;
}

// license key

template<> xlic::license_key from_base64<xlic::license_key>(const std::string & string) {
	xlic::license_key license_key;
	if(xtl::base64::decode(
		license_key.value(), license_key.size(),
		string.c_str(), string.length()
	) != license_key.size()) {
		throw xlic::invalid_license_key(__FILE__, __LINE__);
	}
	return license_key;
}

std::string to_string(const xlic::license_key & license_key) {
	std::string result(28, '\0');
	xtl::base64::encode(&result[0], result.length(), license_key.value(), license_key.size());
	return result;
}

bool license_key::operator<(const xlic::license_key & other) const noexcept {
	return (::memcmp(m_value, other.m_value, sizeof(m_value)) < 0);
}

// signature

template<> xtl::rsa_signature from_base64<xtl::rsa_signature>(const std::string & string) {
	xtl::rsa_signature signature;
	if(xtl::base64::decode(
		signature.value(), signature.size(),
		string.c_str(), string.length()
	) != signature.size()) {
		throw xlic::invalid_signature(__FILE__, __LINE__);
	}
	return signature;
}

std::string to_string(const xtl::rsa_signature & signature) {
	std::string result(88, '\0');
	xtl::base64::encode(&result[0], result.length(), signature.value(), signature.size());
	return result;
}

// serial number

xtl::istream & operator>>(xtl::istream & stream, xlic::serial_number & serial_number) {
	return read(stream, serial_number);
}

xtl::ostream & operator<<(xtl::ostream & stream, const xlic::serial_number & serial_number) {
	return write(stream, serial_number);
}

// license type

void from_string(const std::string & string, xlic::license_type & license_type) {
	if (string == "physical") {
		license_type = xlic::license_type::physical_license;
	} else if (string == "virtual") {
		license_type = xlic::license_type::virtual_license;
	} else {
		license_type = xlic::license_type::unknown;
	}
}

std::string to_string(const xlic::license_type & license_type) {
	switch (license_type) {
		case xlic::license_type::physical_license: { return std::string("physical"); }
		case xlic::license_type::virtual_license: { return std::string("virtual"); }
		default: { return std::string("unknown"); }
	}
}

// attribute list

size_t attribute_list::size() const {
	size_t size = 0;
	for (const auto & attribute : *this) {
		size += sizeof(string_length) + attribute.first.length() +
			sizeof(string_length) + attribute.second.length();
	}
	return size;
}

xtl::istream & operator>>(xtl::istream & stream, xlic::attribute_list & attributes) {
	attributes.clear();
	xlic::array_count count;
	stream >> count;
	std::string first;
	std::string second;
	for (size_t i = count; i > 0; --i) {
		read(stream, first);
		read(stream, second);
		attributes.emplace(std::move(first), std::move(second));
	}
	return stream;
}

xtl::ostream & operator<<(xtl::ostream & stream, const xlic::attribute_list & attributes) {
	stream << xlic::array_count(attributes.count());
	for (const auto & pair : attributes) {
		write(stream, pair.first);
		write(stream, pair.second);
	}
	return stream;
}

// license

size_t license::size() const {
	return sizeof(license_key.size()) +
		sizeof(signature.size()) +
		sizeof(state) +
		sizeof(type) +
		sizeof(valid_after) +
		sizeof(registered_on) +
		sizeof(updated_on) +
		sizeof(grace_start) +
		sizeof(valid_for) +
		sizeof(grace_days) +
		sizeof(nonce) +
		sizeof(string_length) + name.length() +
		sizeof(string_length) + description.length() +
		attributes.size();
}

bool license::operator<(const xlic::license & other) const noexcept {
	return license_key < other.license_key;
}

inline xtl::verify_method & operator+=(xtl::verify_method & method, const xlic::license_key & license_key) {
	return method.append(license_key.value(), license_key.size());
}

inline xtl::verify_method & operator+=(xtl::verify_method & method, const xlic::attribute_list & attributes) {
	for (const auto & pair : attributes) {
		method.append(pair.first.c_str(), pair.first.length());
		method.append(pair.second.c_str(), pair.second.length());
	}
	return method;
}

bool license::verify(const xtl::rsa_public_key & public_key, const xlic::serial_number & serial_number) const {
	xtl::rsa_verify rsa(public_key);
	rsa += to_string(license_key);
	rsa += to_string(type);
	rsa += to_string(valid_after);
	rsa += to_string(registered_on);
	rsa += to_string(updated_on);
	rsa += to_string(valid_for);
	rsa += to_string(grace_days);
	rsa += to_string(nonce);
	rsa += name;
	rsa += attributes;
	rsa += serial_number;
	return rsa.verify(signature);
}

xtl::istream & operator>>(xtl::istream & stream, xlic::license & license) {
	stream >> license.license_key;
	stream >> license.signature;
	stream >> license.state;
	stream >> license.type;
	stream >> license.valid_after;
	stream >> license.registered_on;
	stream >> license.updated_on;
	stream >> license.grace_start;
	stream >> license.valid_for;
	stream >> license.grace_days;
	stream >> license.nonce;
	read(stream, license.name);
	read(stream, license.description);
	stream >> license.attributes;
	return stream;
}

xtl::ostream & operator<<(xtl::ostream & stream, const xlic::license & license) {
	stream << license.license_key;
	stream << license.signature;
	stream << license.state;
	stream << license.type;
	stream << license.valid_after;
	stream << license.registered_on;
	stream << license.updated_on;
	stream << license.grace_start;
	stream << license.valid_for;
	stream << license.grace_days;
	stream << license.nonce;
	write(stream, license.name);
	write(stream, license.description);
	stream << license.attributes;
	return stream;
}

template<> xlic::license from_json<xlic::license>(const xtl::json & js) {
	xlic::license license;
	license.license_key = from_base64<xlic::license_key>(js["license_key"].string());
	license.signature = from_base64<xtl::rsa_signature>(js["signature"].string());
	// license state is determined dynamically
	from_string(js["type"].string(), license.type);
	license.valid_after = js["valid_after"].number();
	license.registered_on = js["registered_on"].number();
	license.updated_on = js["updated_on"].number();
	license.valid_for = js["valid_for"].number();
	license.grace_days = js["grace_days"].number();
	license.nonce = js["nonce"].number();
	license.name = js["name"].string();
	license.description = js["description"].string();
	for (const auto & pair : js["attributes"].object()) {
		license.attributes.emplace(pair.first, pair.second.string());
	}
	return license;
}

// license list

size_t xlic::license_list::size() const {
	size_t size = 0;
	for (const auto & pair : *this) {
		size += pair.second.size();
	}
	return size;
}

xtl::istream & operator>>(xtl::istream & stream, xlic::license_list & licenses) {
	licenses.clear();
	xlic::array_count count;
	stream >> count;
	for(size_t i = count; i > 0; --i) {
		xlic::license license;
		stream >> license;
		licenses.emplace(license.license_key, std::move(license));
	}
	return stream;
}

xtl::ostream & operator<<(xtl::ostream & stream, const xlic::license_list & licenses) {
	stream << xlic::array_count(licenses.count());
	for (const auto & pair : licenses) {
		stream << pair.second;
	}
	return stream;
}

template<> xlic::license_list from_json<xlic::license_list>(const xtl::json & js) {
	xlic::license_list licenses;
	for (const auto & js_license : js.array()) {
		auto license = from_json<xlic::license>(js_license);
		licenses.emplace(license.license_key, std::move(license));
	}
	return licenses;
}

// mac address

inline xtl::hash_method & operator+=(xtl::hash_method & hash_method, const xtl::mac_address & mac_address) {
	hash_method.append(&mac_address, sizeof(mac_address));
	return hash_method;
}

// mac address list

xtl::ostream & operator<<(xtl::ostream & stream, const xlic::mac_address_list & mac_address_list) {
	stream << xlic::array_count(mac_address_list.count());
	for (const xtl::mac_address & mac : mac_address_list) {
		stream << mac;
	}
	return stream;
}

xtl::hash_method & operator+=(xtl::hash_method & hash_method, const xlic::mac_address_list & mac_address_list) {
	for (const xtl::mac_address & mac : mac_address_list) {
		hash_method += mac;
	}
	return hash_method;
}

xtl::istream & operator>>(xtl::istream & stream, xlic::mac_address_list & mac_address_list) {
	mac_address_list.clear();
	xlic::array_count count;
	stream >> count;
	for (size_t i = count; i > 0; --i) {
		xtl::mac_address mac_address;
		stream >> mac_address;
		mac_address_list.emplace(mac_address);
	}
	return stream;
}

// to string

std::string to_string(const time_t & timestamp) {
	std::string result;
	const size_t length = snprintf(nullptr, 0, "%ld", timestamp);
	result.resize(length);
	snprintf(&result[0], length + 1, "%ld", timestamp);
	return result;
}

std::string to_string(const uint32_t & number) {
	std::string result;
	const size_t length = snprintf(nullptr, 0, "%u", number);
	result.resize(length);
	snprintf(&result[0], length + 1, "%u", number);
	return result;
}

std::string to_string(const uint64_t & number) {
	std::string result;
	const size_t length = snprintf(nullptr, 0, "%" PRIu64, number);
	result.resize(length);
	snprintf(&result[0], length + 1, "%" PRIu64, number);
	return result;
}

} // namespace xlic
