#include <xlic/worker.hpp>

#include <xtl/exception.hpp>
#include <xlic/server.hpp>
#include <xlic/session.hpp>
#include <xlic/task.hpp>

#if defined(DEBUG_WORKER) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) printf("worker[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) printf("worker[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xlic {

worker::worker(xlic::server & server)
: m_active(true)
, m_server(server)
, m_acceptor([this] () {
	while (m_active) {
		try {
			auto session = std::make_unique<xlic::session>(*this, m_server.socket().accept());
			m_sessions.emplace(session.get(), std::move(session));
		} catch (const std::exception & e) {
			fdebug("%s\n", e.what());
		}
	}
})
, m_completer([this] () {
	while (m_active) {
		try {
			xlic::task * task = nullptr;
			m_pipe >> task;
			xlic::session * session = task->session();
			if(!task->abandoned()) {
				task->write_result(session->socket());
			}
			const auto it = m_tasks.find(session);
			if (it != m_tasks.end()) {
				it->second.erase(task);
				if(it->second.empty()) {
					m_tasks.erase(it);
				}
			}
		} catch (const std::exception & e) {
			fdebug("%s\n", e.what());
		}
	}
}) {
	fdebug("Created new worker %p", (void *) this);
}

worker::~worker() noexcept {
	fdebug("Destroying worker %p", (void *) this);
}

void worker::close(xlic::session * session) {
	auto it = m_tasks.find(session);
	if (it != m_tasks.end()) {
		for (auto & pair : it->second) {
			pair.second->abandon();
		}
	}
	m_sessions.erase(session);
}

void worker::add(std::unique_ptr<xlic::task> task) {
	m_tasks[task->session()].emplace(task.get(), std::move(task));
}

void worker::completed(xlic::task * task) {
	m_pipe << task;
}

} // namespace xlic
