#ifndef __XTL_FUNCTION__
#define __XTL_FUNCTION__

#include <memory>
#include <tuple>
#include <utility>
#include <xtl/shims.hpp>

namespace xtl {

using std::index_sequence;
using std::get;
using std::make_index_sequence;
using std::tuple;

template<typename> class basic_function;
template<typename, typename...> class function;
template<typename, typename...> class lambda_function;

template<typename return_type, typename... argument_types>
class basic_function<return_type (argument_types...)> {
public:
	inline virtual ~basic_function() noexcept = default;
	virtual return_type operator()(argument_types &&... arguments) = 0;
};

template<typename return_type, typename... argument_types, typename function_type, typename... capture_types>
class lambda_function<return_type (argument_types...), function_type, capture_types...>
	: public basic_function<return_type (argument_types...)>, private tuple<capture_types...> {
public:
	explicit lambda_function(function_type && function, capture_types &&... captures);
	return_type operator()(argument_types &&... arguments) override;

private:
	template<size_t... index>
		return_type invoke(const index_sequence<index...> &, argument_types &&... arguments);

	function_type m_function;
};

template<typename return_type, typename... argument_types>
class function<return_type (argument_types...)> {
public:
	inline function() noexcept = default;
	template<typename function_type, typename... capture_types>
		explicit function(function_type && function, capture_types &&... captures);
	explicit operator bool() const noexcept;
	template<typename... input_types>
		return_type operator()(input_types &&... inputs);

private:
	std::unique_ptr<basic_function<return_type (argument_types...)>> m_function;
};

} // namespace xtl

namespace xtl {

template<typename return_type, typename... argument_types, typename function_type, typename... capture_types>
inline lambda_function<return_type (argument_types...), function_type, capture_types...>::lambda_function(
	function_type && function,
	capture_types &&... captures
)
: tuple<capture_types...>(std::forward<capture_types>(captures)...)
, m_function(std::forward<function_type>(function))
{}

template<typename return_type, typename... argument_types, typename function_type, typename... capture_types>
inline return_type lambda_function<return_type (argument_types...), function_type, capture_types...>::operator()(
	argument_types &&... arguments
) {
	return invoke(
		std::make_index_sequence<sizeof...(capture_types)>(),
		std::forward<argument_types>(arguments)...
	);
}

template<typename return_type, typename... argument_types, typename function_type, typename... capture_types>
template<size_t... index>
inline return_type lambda_function<return_type (argument_types...), function_type, capture_types...>::invoke(
	const index_sequence<index...> &,
	argument_types &&... arguments
) {
	return m_function(
		std::get<index>(*this)...,
		std::forward<argument_types>(arguments)...
	);
}

template<typename return_type, typename... argument_types>
template<typename function_type, typename... capture_types>
inline function<return_type (argument_types...)>::function(
	function_type && function,
	capture_types &&... captures
)
: m_function(new lambda_function<return_type (argument_types...), function_type, capture_types...>(
	std::forward<function_type>(function),
	std::forward<capture_types>(captures)...
))
{}

template<typename return_type, typename... argument_types>
inline function<return_type (argument_types...)>::operator bool() const noexcept {
	return static_cast<bool>(m_function);
}

template<typename return_type, typename... argument_types>
template<typename... input_types>
inline return_type function<return_type (argument_types...)>::operator()(
	input_types &&... inputs
) {
	return m_function->operator()(std::forward<input_types>(inputs)...);
}

} // namespace xtl

#endif
