#ifndef __XTL_JSON__
#define __XTL_JSON__

#include <xtl/exception.hpp>
#include <xtl/istream.hpp>
#include <vector>
#include <string>
#include <map>

namespace xtl {

class json_exception : public xtl::exception {
public:
	inline json_exception(const char * file, int line, const char * reason)
	: exception(file, line, "%s", reason) {}
};

class json {
public:
	using string_type = std::string;
	using object_type = std::map<string_type, json>;
	using array_type = std::vector<json>;
	using number_type = int64_t;

	json() noexcept;
	json(xtl::istream & stream);
	json(string_type && string) noexcept;
	json(number_type && number) noexcept;
	json(bool && boolean) noexcept;
	json(decltype(nullptr)) noexcept;
	json(object_type && object) noexcept;
	json(array_type && array) noexcept;
	json(json && other) noexcept;
	json & operator=(json && other) noexcept;
	~json() noexcept;

	size_t size() const noexcept; // object / array size, string length
	const json & operator[](const size_t index) const; // array
	const json & operator[](const string_type & key) const; // object
	const json & operator[](const char * key) const; // object

	bool is_string() const noexcept;
	bool is_number() const noexcept;
	bool is_boolean() const noexcept;
	bool is_null() const noexcept;
	bool is_object() const noexcept;
	bool is_array() const noexcept;

	string_type string() const;
	number_type number() const;
	bool boolean() const;
	const object_type & object() const;
	const array_type & array() const;

private:
	enum class type { unknown, string, number, boolean, null, object, array };

	type m_type;
	union {
		string_type m_string;
		number_type m_number;
		bool m_boolean;
		object_type m_object;
		array_type m_array;
	};
};

inline json::json() noexcept : m_type(type::unknown) {}
inline json::json(string_type && string) noexcept : m_type(type::string), m_string(std::move(string)) {}
inline json::json(number_type && number) noexcept : m_type(type::number), m_number(std::move(number)) {}
inline json::json(bool && boolean) noexcept : m_type(type::boolean), m_boolean(boolean) {}
inline json::json(decltype(nullptr)) noexcept : m_type(type::null) {}
inline json::json(object_type && object) noexcept : m_type(type::object), m_object(std::move(object)) {}
inline json::json(array_type && array) noexcept : m_type(type::array), m_array(std::move(array)) {}

inline bool json::is_string() const noexcept {
	return m_type == type::string;
}

inline bool json::is_number() const noexcept {
	return m_type == type::number;
}

inline bool json::is_boolean() const noexcept {
	return m_type == type::boolean;
}

inline bool json::is_null() const noexcept {
	return m_type == type::null;
}

inline bool json::is_object() const noexcept {
	return m_type == type::object;
}

inline bool json::is_array() const noexcept {
	return m_type == type::array;
}

} // namespace json

#endif
