#include <xtl/base64.hpp>

#include <cstdint>
#include <cstdio>
#include <algorithm>

namespace xtl {

namespace base64 {

size_t encoded_size(
	size_t input_size
) {
	switch (input_size % 3) {
		case 2: return (((input_size + 1) / 3) * 4);
		case 1: return (((input_size + 2) / 3) * 4);
		default: return (input_size / 3) * 4;
	}
}

size_t decoded_size(
	const void * input,
	size_t input_size
) {
	const char * in = static_cast<const char *>(input);
	if (input_size > 0 && in[input_size - 1] == '=') {
		if (input_size > 1 && in[input_size - 2] == '=') {
			return ((input_size / 4) * 3) - 2;
		}
		return ((input_size / 4) * 3) - 1;
	}
	return (input_size / 4) * 3;
}

size_t decode(
	void * output,
	size_t output_size,
	const void * input,
	size_t input_size
) {
	// This maps provides efficient conversion between the ascii and base64 character set.
	// Conversion is done by subtracting 43 from characters and performing index loopups.
	static const uint8_t map[128] = {
		62, // "+"
		0, 0, 0,
		63, // "/"
		52, 53, 54, 55, 56, 57, 58, 59, 60, 61, // 0-9
		0, 0, 0, 0, 0, 0, 0,
		0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, // A-Z
		0, 0, 0, 0, 0, 0,
		26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, // a-z
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0,
	};
	const size_t input_blocks = input_size / 4;
	const size_t output_blocks = output_size / 3;
	const size_t rounds = std::min(input_blocks, output_blocks);
	uint8_t * out = static_cast<uint8_t *>(output);
	const char * in = static_cast<const char *>(input);
	for (size_t i = rounds; i > 0; --i) {
		const uint8_t c1 = map[((*in++) - 43) & 0x7f];
		const uint8_t c2 = map[((*in++) - 43) & 0x7f];
		const uint8_t c3 = map[((*in++) - 43) & 0x7f];
		const uint8_t c4 = map[((*in++) - 43) & 0x7f];
		const uint32_t block = (c1 << 18) | ( c2 << 12) | (c3 << 6) | (c4);
		*out++ = (block >> 16) & 0xff;
		*out++ = (block >> 8) & 0xff;
		*out++ = block & 0xff;
	}
	if (rounds < input_blocks) {
		const uint8_t c1 = map[((*in++) - 43) & 0x7f];
		const uint8_t c2 = map[((*in++) - 43) & 0x7f];
		const uint8_t c3 = map[((*in++) - 43) & 0x7f];
		const uint8_t c4 = map[((*in++) - 43) & 0x7f];
		const uint32_t block = (c1 << 18) | ( c2 << 12) | (c3 << 6) | (c4);
		switch (output_size % 3) {
			case 2: {
				*out++ = (block >> 16) & 0xff;
				*out++ = (block >> 8) & 0xff;
				break;
			}
			case 1: {
				*out++ = (block >> 16) & 0xff;
				break;
			}
		}
	}
	return decoded_size(input, input_size);
}

size_t encode(
	void * output,
	size_t output_size,
	const void * input,
	size_t input_size
) {
	// This maps provides efficient conversion between the base64 and ascii character set.
	static const char map[64] = {
		'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
		'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
		'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
		'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/'
	};
	const size_t input_blocks = input_size / 3;
	const size_t output_blocks = output_size / 4;
	const size_t rounds = std::min(input_blocks, output_blocks);
	const uint8_t * in = static_cast<const uint8_t *>(input);
	char * out = static_cast<char *>(output);
	for (size_t i = rounds; i > 0; --i) {
		const uint8_t b1 = *in++;
		const uint8_t b2 = *in++;
		const uint8_t b3 = *in++;
		const uint32_t block = b1 << 16 | b2 << 8 | b3;
		*out++ = map[(block >> 18) & 0x3f];
		*out++ = map[(block >> 12) & 0x3f];
		*out++ = map[(block >> 6) & 0x3f];
		*out++ = map[block & 0x3f];
	}
	if (rounds < output_blocks) {
		switch (input_size % 3) {
			case 2: {
				const uint8_t b1 = *in++;
				const uint8_t b2 = *in++;
				const uint32_t block = b1 << 16 | b2 << 8;
				*out++ = map[(block >> 18) & 0x3f];
				*out++ = map[(block >> 12) & 0x3f];
				*out++ = map[(block >> 6) & 0x3f];
				*out++ = '=';
				break;
			}
			case 1: {
				const uint8_t b1 = *in++;
				const uint32_t block = b1 << 16;
				*out++ = map[(block >> 18) & 0x3f];
				*out++ = map[(block >> 12) & 0x3f];
				*out++ = '=';
				*out++ = '=';
				break;
			}
		}
	}
	return encoded_size(input_size);
}

} // namespace base64

} // namespace xtl
