#include <xtl/cryptostream.hpp>

#include <xtl/exception.hpp>
#include <sodium.h>

#include <cstdint>
#include <cstdlib>
#include <vector>
#include <cstring>

#if defined(DEBUG_CRYPTOSTREAM) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) fprintf(stderr, "cryptostream[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) fprintf(stderr, "cryptostream[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

#ifndef crypto_stream_BLOCKSIZE
#define crypto_stream_BLOCKSIZE 64
#endif

namespace xtl {

namespace crypto {

size_t block_size() {
	return crypto_stream_BLOCKSIZE;
}

size_t key_size() {
	return crypto_stream_KEYBYTES;
}

size_t iv_size() {
	return crypto_stream_NONCEBYTES;
}

void randomize(void * buffer, size_t amount) {
	randombytes_buf(buffer, amount);
}

} // crypto

class cryptostream_impl {
public:
	inline cryptostream_impl(xtl::iostream & stream, size_t buffer_size)
	: m_stream(stream)
	, m_rx_blocks(0), m_tx_blocks(0), m_buffer_size(buffer_size)
	, m_rx_base(0), m_rx_plaintext(0), m_rx_encrypted(0)
	, m_tx_base(0), m_tx_encrypted(0), m_tx_plaintext(0)
	, m_rx_buffer(m_buffer_size), m_tx_buffer(m_buffer_size)
	{
		memset(m_rx_key, 0, sizeof(m_rx_key));
		memset(m_tx_key, 0, sizeof(m_tx_key));
		memset(m_rx_iv, 0, sizeof(m_rx_iv));
		memset(m_tx_iv, 0, sizeof(m_tx_iv));
	}

	inline size_t read(void * buffer, size_t amount) {
		for (size_t remaining = amount; remaining > 0;) {
			const size_t offset = amount - remaining;
			remaining -= drain_input(static_cast<uint8_t *>(buffer) + offset, remaining);
			if (remaining > 0) {
				if (m_rx_encrypted < crypto_stream_BLOCKSIZE) {
					if (read_stream() == 0) {
						return amount - remaining;
					}
				}
				while (m_rx_encrypted >= crypto_stream_BLOCKSIZE) {
					decrypt_input_block();
				}
			}
		}
		return amount;
	}

	inline size_t write(const void * buffer, size_t amount) {
		const size_t block_size = crypto_stream_BLOCKSIZE;
		const size_t buffer_size = m_buffer_size;
		for (size_t remaining = amount; remaining > 0;) {
			const size_t offset = amount - remaining;
			remaining -= fill_output(static_cast<const uint8_t *>(buffer) + offset, remaining);
			const size_t used = m_tx_encrypted + m_tx_plaintext;
			if (buffer_size == used) {
				while (m_tx_plaintext >= block_size) {
					encrypt_output_block();
				}
				if (m_tx_encrypted > 0) {
					if (write_stream() == 0) {
						return amount - remaining;
					}
				}
			}
		}
		return amount;
	}

	inline size_t flush() {
		const size_t buffer_size = m_buffer_size;
		const size_t plaintext_size = m_tx_plaintext;
		const size_t remainder = plaintext_size % crypto_stream_BLOCKSIZE;
		if (remainder > 0) {
			const size_t padding_size = crypto_stream_BLOCKSIZE - remainder;
			const size_t offset = (m_tx_base + m_tx_encrypted + plaintext_size) % buffer_size;
			fdebug("Padding stream with %zu bytes to %zu", padding_size, offset + padding_size);
			randombytes_buf(&m_tx_buffer[offset], padding_size);
			m_tx_plaintext += padding_size;
		}
		while (m_tx_plaintext >= crypto_stream_BLOCKSIZE) {
			encrypt_output_block();
		}
		size_t result = m_tx_encrypted;
		while (m_tx_encrypted > 0) {
			write_stream();
		}
		return result;
	}

	inline void set_read_key(const void * buffer, size_t amount) {
		if (amount != crypto_stream_KEYBYTES) {
			throw invalid_argument(__FILE__, __LINE__);
		}
		::memcpy(m_rx_key, buffer, crypto_stream_KEYBYTES);
	}

	inline void set_write_key(const void * buffer, size_t amount) {
		if (amount != crypto_stream_KEYBYTES) {
			throw invalid_argument(__FILE__, __LINE__);
		}
		::memcpy(m_tx_key, buffer, crypto_stream_KEYBYTES);
	}

	inline void set_read_iv(const void * buffer, size_t amount) {
		if (amount != crypto_stream_NONCEBYTES) {
			throw invalid_argument(__FILE__, __LINE__);
		}
		::memcpy(m_rx_iv, buffer, crypto_stream_NONCEBYTES);
	}

	inline void set_write_iv(const void * buffer, size_t amount) {
		if (amount != crypto_stream_NONCEBYTES) {
			throw invalid_argument(__FILE__, __LINE__);
		}
		::memcpy(m_tx_iv, buffer, crypto_stream_NONCEBYTES);
	}

private:
	xtl::iostream & m_stream;
	size_t m_rx_blocks, m_tx_blocks, m_buffer_size;
	size_t m_rx_base, m_rx_plaintext, m_rx_encrypted;
	size_t m_tx_base, m_tx_encrypted, m_tx_plaintext;
	std::vector<uint8_t> m_rx_buffer, m_tx_buffer;
	uint8_t m_rx_key[crypto_stream_KEYBYTES], m_tx_key[crypto_stream_KEYBYTES];
	uint8_t m_rx_iv[crypto_stream_NONCEBYTES], m_tx_iv[crypto_stream_NONCEBYTES];

	size_t drain_input(void * buffer, size_t amount) {
		const size_t plaintext_size = m_rx_plaintext;
		fdebug("drain_input(%p, %zu), plaintext = %zu", buffer, amount, plaintext_size);
		if (plaintext_size > 0) {
			amount = std::min(plaintext_size, amount);
			const size_t buffer_size = m_buffer_size;
			const size_t plaintext_offset = m_rx_base;
			const size_t max = std::min(plaintext_offset + amount, buffer_size) - plaintext_offset;
			::memcpy(buffer, &m_rx_buffer[plaintext_offset], max);
			m_rx_base = (m_rx_base + max) % buffer_size;
			m_rx_plaintext -= max;
			return max;
		}
		return 0;
	}

	size_t read_stream() {
		const size_t buffer_size = m_buffer_size;
		const size_t plaintext_size = m_rx_plaintext;
		const size_t encrypted_size = m_rx_encrypted;
		const size_t unused_offset = (m_rx_base + plaintext_size + encrypted_size) % buffer_size;
		const size_t unused_size = buffer_size - plaintext_size - encrypted_size;
		const size_t max = std::min(unused_offset + unused_size, buffer_size) - unused_offset;
		fdebug("read_stream(), unused = %zu@%zu, max = %zu", unused_size, unused_offset, max);
		const size_t result = m_stream.read(&m_rx_buffer[unused_offset], max);
		m_rx_encrypted += result;
		return result;
	}

	void decrypt_input_block() {
		const size_t block_size = crypto_stream_BLOCKSIZE;
		const size_t buffer_size = m_buffer_size;
		const size_t plaintext_offset = m_rx_base;
		const size_t plaintext_size = m_rx_plaintext;
		const size_t encrypted_offset = (plaintext_offset + plaintext_size) % buffer_size;
		if (encrypted_offset + block_size > buffer_size) {
			// wrapping
			uint8_t tmp[block_size];
			const size_t second_half = (encrypted_offset + block_size) % buffer_size;
			const size_t first_half = block_size - second_half;
			fdebug(
				"decrypt_input_block() (wrapping) %zu-%zu + %zu-%zu",
				encrypted_offset, encrypted_offset + first_half,
				size_t(0), second_half
			);
			::memcpy(tmp, &m_rx_buffer[encrypted_offset], first_half);
			::memcpy(&tmp[first_half], &m_rx_buffer[0], second_half);
			if (crypto_stream_xsalsa20_xor_ic(tmp, tmp, block_size, m_rx_iv, m_rx_blocks++, m_rx_key) != 0) {
				throw xtl::invalid_argument(__FILE__, __LINE__);
			}
			::memcpy(&m_rx_buffer[encrypted_offset], tmp, first_half);
			::memcpy(&m_rx_buffer[0], &tmp[first_half], second_half);
		} else {
			// contiguous
			fdebug(
				"decrypt_input_block() (contiguous) %zu-%zu",
				encrypted_offset, encrypted_offset + block_size
			);
			uint8_t * ptr = &m_rx_buffer[encrypted_offset];
			if (crypto_stream_xsalsa20_xor_ic(ptr, ptr, block_size, m_rx_iv, m_rx_blocks++, m_rx_key) != 0) {
				throw xtl::invalid_argument(__FILE__, __LINE__);
			}
		}
		m_rx_plaintext += block_size;
		m_rx_encrypted -= block_size;
	}

	size_t fill_output(const void * buffer, size_t amount) {
		const size_t buffer_size = m_buffer_size;
		const size_t encrypted_size = m_tx_encrypted;
		const size_t plaintext_size = m_tx_plaintext;
		const size_t unused_size = buffer_size - encrypted_size - plaintext_size;
		fdebug("fill_output(%p, %zu), unused = %zu", buffer, amount, unused_size);
		if (unused_size > 0) {
			amount = std::min(unused_size, amount);
			const size_t unused_offset = (m_tx_base + encrypted_size + plaintext_size) % buffer_size;
			const size_t max = std::min(unused_offset + amount, buffer_size) - unused_offset;
			::memcpy(&m_tx_buffer[unused_offset], buffer, max);
			m_tx_plaintext += max;
			return max;
		}
		return 0;
	}

	void encrypt_output_block() {
		const size_t block_size = crypto_stream_BLOCKSIZE;
		const size_t buffer_size = m_buffer_size;
		const size_t encrypted_size = m_tx_encrypted;
		const size_t encrypted_offset = m_tx_base;
		const size_t plaintext_offset = (encrypted_offset + encrypted_size) % buffer_size;
		if (plaintext_offset + block_size > buffer_size) {
			// wrapping
			uint8_t tmp[block_size];
			const size_t second_half = (plaintext_offset + block_size) % buffer_size;
			const size_t first_half = block_size - second_half;
			fdebug(
				"encrypt_output_block() (wrapping) %zu-%zu + %zu-%zu",
				plaintext_offset, plaintext_offset + first_half,
				size_t(0), second_half
			);
			::memcpy(tmp, &m_tx_buffer[plaintext_offset], first_half);
			::memcpy(&tmp[first_half], &m_tx_buffer[0], second_half);
			if (crypto_stream_xsalsa20_xor_ic(tmp, tmp, block_size, m_tx_iv, m_tx_blocks++, m_tx_key) != 0) {
				throw xtl::invalid_argument(__FILE__, __LINE__);
			}
		} else {
			// contiguous
			fdebug(
				"encrypt_output_block() (contiguous) %zu-%zu",
				plaintext_offset, plaintext_offset + block_size
			);
			uint8_t * ptr = &m_tx_buffer[plaintext_offset];
			if (crypto_stream_xsalsa20_xor_ic(ptr, ptr, block_size, m_tx_iv, m_tx_blocks++, m_tx_key) != 0) {
				throw xtl::invalid_argument(__FILE__, __LINE__);
			}
		}
		m_tx_plaintext -= block_size;
		m_tx_encrypted += block_size;
	}

	size_t write_stream() {
		const size_t buffer_size = m_buffer_size;
		const size_t encrypted_offset = m_tx_base;
		const size_t encrypted_size = m_tx_encrypted;
		const size_t max = std::min(encrypted_offset + encrypted_size, buffer_size) - encrypted_offset;
		fdebug("write_stream(), encrypted = %zu@%zu, max = %zu", encrypted_size, encrypted_offset, max);
		const size_t result = m_stream.write(&m_tx_buffer[encrypted_offset], max);
		m_tx_base = (m_tx_base + result) % buffer_size;
		m_tx_encrypted -= result;
		return result;
	}
};

cryptostream::cryptostream(xtl::iostream & stream, size_t buffer_size)
: shared_ptr(std::make_shared<xtl::cryptostream_impl>(stream, buffer_size))
{}

void cryptostream::set_read_key(const void * buffer, size_t amount) {
	get()->set_read_key(buffer, amount);
}

void cryptostream::set_write_key(const void * buffer, size_t amount) {
	get()->set_write_key(buffer, amount);
}

void cryptostream::set_read_iv(const void * buffer, size_t amount) {
	get()->set_read_iv(buffer, amount);
}

void cryptostream::set_write_iv(const void * buffer, size_t amount) {
	get()->set_write_iv(buffer, amount);
}

size_t cryptostream::read(void * buffer, size_t amount) {
	return get()->read(buffer, amount);
}

size_t cryptostream::write(const void * buffer, size_t amount) {
	return get()->write(buffer, amount);
}

size_t cryptostream::flush() {
	return get()->flush();
}

} // namespace xtl
