#include <xtl/filestream.hpp>

#include <xtl/exception.hpp>
#include <xtl/file_descriptor.hpp>

#include <cerrno>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#if defined(DEBUG_FILESTREAM) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) fprintf(stderr, "filestream[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) fprintf(stderr, "filestream[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xtl {

inline int make_file(const std::string & filename, int flags) {
	const int fd = ::open(filename.c_str(), flags, 00666);
	if (fd < 0) {
		switch (errno) {
			case EACCES: throw xtl::access_denied(__FILE__, __LINE__);
			case EDQUOT: throw xtl::quota_reached(__FILE__, __LINE__);
			case EEXIST: throw xtl::file_already_exists(__FILE__, __LINE__);
			case EINVAL: throw xtl::invalid_argument(__FILE__, __LINE__);
			case EISDIR: throw xtl::invalid_argument(__FILE__, __LINE__);
			case ENAMETOOLONG: throw xtl::name_too_long(__FILE__, __LINE__);
			case ENOENT: throw xtl::file_not_found(__FILE__, __LINE__);
			case ENOTDIR: throw xtl::invalid_argument(__FILE__, __LINE__);
			case EPERM: throw xtl::access_denied(__FILE__, __LINE__);
			case EROFS: throw xtl::access_denied(__FILE__, __LINE__);
			case EBADF: throw xtl::invalid_argument(__FILE__, __LINE__);

		}
		throw xtl::system_exception(__FILE__, __LINE__);
	}
	return fd;
}

inline int convert(xtl::filemode mode) {
	switch (mode) {
		case xtl::filemode::readonly: return O_RDONLY;
		case xtl::filemode::writeonly: return O_WRONLY | O_CREAT;
		case xtl::filemode::readwrite: return O_RDWR | O_CREAT;
	}
	throw xtl::invalid_argument(__FILE__, __LINE__);
}

class filestream_impl {
public:
	inline filestream_impl(const std::string & filename, xtl::filemode mode)
	: m_fd(make_file(filename, convert(mode) | O_CLOEXEC | O_LARGEFILE))
	, m_filename(filename)
	{}

	inline size_t read(void * buffer, size_t amount) {
		if (amount > 0) {
			for (;;) {
				const ssize_t result = ::read(m_fd.value(), buffer, amount);
				if (result > 0) {
					return result;
				} else if (result == 0) {
					throw xtl::end_of_stream(__FILE__, __LINE__);
				} else if (errno == EINTR) {
					continue;
				} else {
					throw xtl::system_exception(__FILE__, __LINE__);
				}
			}
		}
		return 0;
	}

	inline size_t write(const void * buffer, size_t amount) {
		if (amount > 0) {
			for (;;) {
				const ssize_t result = ::write(m_fd.value(), buffer, amount);
				if (result >= 0) {
					return result;
				} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
					return 0;
				} else if (errno == EINTR) {
					continue;
				} else {
					throw xtl::system_exception(__FILE__, __LINE__);
				}
			}
		}
		return 0;
	}

	inline size_t flush() {
		::fsync(m_fd.value());
		return 0;
	}

	inline size_t size() const {
		struct stat buffer;
		if (::fstat(m_fd.value(), &buffer) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
		return buffer.st_size;
	}

	inline void chmod(int mode) {
		if (::fchmod(m_fd.value(), mode) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
	}

	inline void truncate(size_t size) {
		for (;;) {
			if (::ftruncate(m_fd.value(), size) != 0) {
				throw xtl::system_exception(__FILE__, __LINE__);
			}
			return;
		}
	}

private:
	xtl::file_descriptor m_fd;
	std::string m_filename;
};

filestream::filestream(const std::string & filename, xtl::filemode mode)
: shared_ptr(std::make_shared<xtl::filestream_impl>(filename, mode))
{}

size_t filestream::size() const {
	return get()->size();
}

void filestream::chmod(int mode) {
	return get()->chmod(mode);
}

void filestream::truncate(size_t size) {
	return get()->truncate(size);
}

size_t filestream::read(void * buffer, size_t amount) {
	return get()->read(buffer, amount);
}

size_t filestream::write(const void * buffer, size_t amount) {
	return get()->write(buffer, amount);
}

size_t filestream::flush() {
	return get()->flush();
}

} // namespace xtl
