#include <xtl/https_stream.hpp>

#include <xtl/exception.hpp>
#include <curl/curl.h>
#include <unordered_map>
#include <vector>
#include <cstring>

#if defined(DEBUG_HTTPS_STREAM) || defined(DEBUG_ALL)
#define HTTPS_STREAM_VERBOSE true
#define fdebug(fmt, ...) fprintf(stderr, "https_stream[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) fprintf(stderr, "https_stream[%d]: " msg "\n", __LINE__);
#else
#define HTTPS_STREAM_VERBOSE false
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xtl {

inline size_t multiple_of(size_t value, size_t multiple) noexcept {
	const size_t remainder = value % multiple;
	if (remainder > 0) {
		return value + (multiple - remainder);
	}
	return value;
}

class curl_easy_exception : public xtl::https_stream_exception {
public:
	inline curl_easy_exception(const char * file, int line, CURLcode code)
	: https_stream_exception(file, line, ::curl_easy_strerror(code)) {}
};

class curl_slist {
public:
	inline constexpr curl_slist() noexcept
	: m_list(nullptr) {}

	inline ~curl_slist() noexcept {
		if (m_list) {
			::curl_slist_free_all(m_list);
		}
	}

	inline ::curl_slist * get() const noexcept {
		return m_list;
	}

	void append(const std::string & value) {
		m_list = curl_slist_append(m_list, value.c_str());
		if (m_list == nullptr) {
			throw xtl::system_exception(__FILE__, __LINE__, ENOMEM);
		}
	}

private:
	struct ::curl_slist * m_list;
};

class curl {
public:
	typedef size_t (* callback_type)(char * buffer, size_t size, size_t count, void * ptr);

	inline curl()
	: m_handle(::curl_easy_init())
	{
		if (m_handle == nullptr) {
			throw xtl::curl_easy_exception(__FILE__, __LINE__, CURLE_FAILED_INIT);
		}
	}

	inline ~curl() noexcept {
		::curl_easy_cleanup(m_handle);
	}

	void set_option(CURLoption option, long value) {
		const auto result = ::curl_easy_setopt(m_handle, option, value);
		if (result != CURLE_OK) {
			throw xtl::curl_easy_exception(__FILE__, __LINE__, result);
		}
	}

	void set_option(CURLoption option, const std::string & value) {
		const auto result = ::curl_easy_setopt(m_handle, option, value.c_str());
		if (result != CURLE_OK) {
			throw xtl::curl_easy_exception(__FILE__, __LINE__, result);
		}
	}

	void set_option(CURLoption option, void * ptr) {
		const auto result = ::curl_easy_setopt(m_handle, option, ptr);
		if (result != CURLE_OK) {
			throw xtl::curl_easy_exception(__FILE__, __LINE__, result);
		}
	}

	void set_option(CURLoption option, callback_type callback) {
		const auto result = ::curl_easy_setopt(m_handle, option, callback);
		if (result != CURLE_OK) {
			throw xtl::curl_easy_exception(__FILE__, __LINE__, result);
		}
	}

	void set_option(CURLoption option, const xtl::curl_slist & slist) {
		const auto result = ::curl_easy_setopt(m_handle, option, slist.get());
		if (result != CURLE_OK) {
			throw xtl::curl_easy_exception(__FILE__, __LINE__, result);
		}
	}

	void get_info(CURLINFO info, long & value) const {
		const auto result = ::curl_easy_getinfo(m_handle, info, &value);
		if (result != CURLE_OK) {
			throw xtl::curl_easy_exception(__FILE__, __LINE__, result);
		}
	}

	void perform() {
		const auto result = ::curl_easy_perform(m_handle);
		if (result != CURLE_OK) {
			throw xtl::curl_easy_exception(__FILE__, __LINE__, result);
		}
	}

private:
	CURL * m_handle;
};

class https_stream_impl {
public:
	inline https_stream_impl(const std::string & hostname, uint16_t port)
	: m_hostname(hostname)
	{
		m_curl.set_option(CURLOPT_VERBOSE, long(HTTPS_STREAM_VERBOSE));
		m_curl.set_option(CURLOPT_NOPROGRESS, long(true));
		m_curl.set_option(CURLOPT_PORT, long(port));
		m_curl.set_option(CURLOPT_HEADERFUNCTION, receive_headers);
		m_curl.set_option(CURLOPT_HEADERDATA, this);
		m_curl.set_option(CURLOPT_READFUNCTION, send_body);
		m_curl.set_option(CURLOPT_READDATA, this);
		m_curl.set_option(CURLOPT_WRITEFUNCTION, receive_body);
		m_curl.set_option(CURLOPT_WRITEDATA, this);
		m_curl.set_option(CURLOPT_FOLLOWLOCATION, long(true));
		m_curl.set_option(CURLOPT_POSTREDIR, CURL_REDIR_POST_ALL);
		// Disable SSL verification
		m_curl.set_option(CURLOPT_SSL_VERIFYHOST, long(false));
		m_curl.set_option(CURLOPT_SSL_VERIFYPEER, long(false));
		#ifdef CURLOPT_SSL_VERIFYSTATUS
			m_curl.set_option(CURLOPT_SSL_VERIFYSTATUS, long(false));
		#endif
	}

	inline void get(const std::string & url) {
		reset();
		m_curl.set_option(CURLOPT_URL, "https://" + m_hostname + url);
		m_curl.set_option(CURLOPT_HTTPGET, long(true));
	}

	inline void post(const std::string & url) {
		reset();
		m_curl.set_option(CURLOPT_URL, "https://" + m_hostname + url);
		m_curl.set_option(CURLOPT_POST, long(true));
	}

	inline void put(const std::string & url) {
		reset();
		m_curl.set_option(CURLOPT_URL, "https://" + m_hostname + url);
		m_curl.set_option(CURLOPT_UPLOAD, long(true));
		// Disable Expect: 100-continue"
		unset_header("Expect");
	}

	inline void set_timeout(size_t timeout_ms) {
		m_curl.set_option(CURLOPT_TIMEOUT_MS, timeout_ms);
	}

	inline void set_header(const std::string & name, const std::string & value) {
		m_request_headers[name] = value;
	}

	inline void unset_header(const std::string & name) {
		m_request_headers[name] = "";
	}

	inline void referer(const std::string & value) {
		m_curl.set_option(CURLOPT_REFERER, value);
	}

	inline void useragent(const std::string & value) {
		m_curl.set_option(CURLOPT_USERAGENT, value);
	}

	inline void content_type(const std::string & value) {
		m_request_headers["Content-Type"] = value;
	}

	inline int code() const {
		long code = 0;
		m_curl.get_info(CURLINFO_RESPONSE_CODE, code);
		return code;
	}

	inline size_t read(void * buffer, size_t amount) {
		const size_t size = m_rx_body_offset;
		const size_t offset = m_read_offset;
		const size_t max = std::min(amount, size - offset);
		if (max > 0) {
			::memcpy(buffer, &m_response_body[offset], max);
			m_read_offset += max;
			return max;
		}
		throw xtl::end_of_stream(__FILE__, __LINE__);
	}

	inline size_t write(const void * buffer, size_t amount) {
		if (amount > 0) {
			const size_t offset = m_write_size;
			m_request_body.resize(multiple_of(offset + amount, 4096));
			::memcpy(&m_request_body[offset], buffer, amount);
			m_write_size += amount;
			return amount;
		}
		return 0;
	}

	inline size_t flush() {
		m_curl.set_option(CURLOPT_INFILESIZE, m_write_size);
		m_curl.set_option(CURLOPT_POSTFIELDSIZE, m_write_size);
		xtl::curl_slist headers;
		for (const auto & pair : m_request_headers) {
			if (pair.second.length() > 0) {
				headers.append(pair.first + ": " + pair.second);
			} else {
				headers.append(pair.first + ":");
			}
		}
		m_curl.set_option(CURLOPT_HTTPHEADER, headers);
		m_rx_body_offset = 0;
		m_curl.perform();
		m_tx_body_offset = 0;
		fdebug("body: %.*s\n", int(m_rx_body_offset), m_response_body.data());
		return m_write_size;
	}

private:
	xtl::curl m_curl;
	std::string m_hostname;
	size_t m_read_offset;
	size_t m_write_size;
	size_t m_rx_header_offset;
	size_t m_rx_body_offset;
	size_t m_tx_body_offset;
	std::unordered_map<std::string, std::string> m_request_headers;
	std::vector<char> m_response_headers;
	std::vector<char> m_request_body;
	std::vector<char> m_response_body;

	void reset() {
		m_request_headers.clear();
		m_response_headers.clear();
		m_request_body.clear();
		m_response_body.clear();
		m_read_offset = 0;
		m_write_size = 0;
		m_rx_header_offset = 0;
		m_rx_body_offset = 0;
		m_tx_body_offset = 0;
	}

	static size_t receive_headers(char * buffer, size_t size, size_t count, void * ptr) {
		auto * self = static_cast<xtl::https_stream_impl *>(ptr);
		const size_t amount = size * count;
		fdebug("receive_headers(%zu)", amount);
		const size_t offset = self->m_rx_header_offset;
		self->m_response_headers.resize(multiple_of(offset + amount, 4096));
		::memcpy(&self->m_response_headers[offset], buffer, amount);
		self->m_rx_header_offset += amount;
		return amount;
	}

	static size_t receive_body(char * buffer, size_t size, size_t count, void * ptr) {
		auto * self = static_cast<xtl::https_stream_impl *>(ptr);
		const size_t amount = size * count;
		fdebug("receive_body(%zu)", amount);
		const size_t offset = self->m_rx_body_offset;
		self->m_response_body.resize(multiple_of(offset + amount, 4096));
		::memcpy(&self->m_response_body[offset], buffer, amount);
		self->m_rx_body_offset += amount;
		return amount;
	}

	static size_t send_body(char * buffer, size_t size, size_t count, void * ptr) {
		auto * self = static_cast<xtl::https_stream_impl *>(ptr);
		const size_t amount = size * count;
		const size_t offset = self->m_tx_body_offset;
		const size_t max = std::min(amount, self->m_write_size - offset);
		if (max > 0) {
			::memcpy(buffer, &self->m_request_body[offset], max);
		}
		self->m_tx_body_offset += max;
		fdebug("send_body(%zu) = %zu", amount, max);
		return max;
	}
};

https_stream::https_stream(const std::string & hostname, uint16_t port)
: shared_ptr(std::make_shared<xtl::https_stream_impl>(hostname, port))
{}

void https_stream::get(const std::string & url) {
	shared_ptr::get()->get(url);
}

void https_stream::post(const std::string & url) {
	shared_ptr::get()->post(url);
}

void https_stream::put(const std::string & url) {
	shared_ptr::get()->put(url);
}

void https_stream::set_timeout(size_t timeout_ms) {
	shared_ptr::get()->set_timeout(timeout_ms);
}

void https_stream::set_header(const std::string & name, const std::string & value) {
	shared_ptr::get()->set_header(name, value);
}

void https_stream::unset_header(const std::string & name) {
	shared_ptr::get()->unset_header(name);
}

void https_stream::referer(const std::string & value) {
	shared_ptr::get()->referer(value);
}

void https_stream::useragent(const std::string & value) {
	shared_ptr::get()->useragent(value);
}

void https_stream::content_type(const std::string & value) {
	shared_ptr::get()->content_type(value);
}

int https_stream::code() const {
	return shared_ptr::get()->code();
}

size_t https_stream::read(void * buffer, size_t amount) {
	return shared_ptr::get()->read(buffer, amount);
}

size_t https_stream::write(const void * buffer, size_t amount) {
	return shared_ptr::get()->write(buffer, amount);
}

size_t https_stream::flush() {
	return shared_ptr::get()->flush();
}

} // namespace xtl
