#include <xtl/interval_timer.hpp>

#include <xtl/exception.hpp>
#include <xtl/file_descriptor.hpp>
#include <xtl/scheduler.hpp>
#include <sys/timerfd.h>
#include <unistd.h>

#if defined(DEBUG_interval_timer) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) fprintf(stderr, "interval_timer[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) fprintf(stderr, "interval_timer[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xtl {

inline int make_interval_timer() {
	const int fd = timerfd_create(CLOCK_MONOTONIC, TFD_NONBLOCK|TFD_CLOEXEC);
	if (fd < 0) {
		throw xtl::system_exception(__FILE__, __LINE__);
	}
	return fd;
}

class interval_timer_impl {
public:
	inline interval_timer_impl(size_t interval_ms)
	: m_fd(make_interval_timer())
	{
		struct itimerspec ts;
		const time_t interval_seconds = interval_ms / 1000;
		const long interval_nseconds = (interval_ms % 1000) * 1000000;
		ts.it_interval.tv_sec = interval_seconds;
		ts.it_interval.tv_nsec = interval_nseconds;
		ts.it_value.tv_sec = interval_seconds;
		ts.it_value.tv_nsec = interval_nseconds;
		if (timerfd_settime(m_fd.value(), 0, &ts, nullptr) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
		fdebug("Created new interval_timer %p", (void *) this);
		xtl::scheduler::monitor(m_fd);
	}

	inline ~interval_timer_impl() noexcept {
		fdebug("Destroying interval_timer %p", (void *) this);
		xtl::scheduler::release(m_fd);
	}

	inline void wait() {
		for (;;) {
			uint64_t dummy;
			const ssize_t result = ::read(m_fd.value(), &dummy, sizeof(dummy));
			if (result > 0) {
				return;
			} else if (result == 0) {
				throw xtl::end_of_stream(__FILE__, __LINE__);
			} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
				xtl::scheduler::wait_for(m_fd);
			} else if (errno == EINTR) {
				continue;
			} else {
				throw xtl::system_exception(__FILE__, __LINE__);
			}
		}
	}

private:
	xtl::file_descriptor m_fd;
};

interval_timer::interval_timer(size_t interval_ms)
: shared_ptr(std::make_shared<interval_timer_impl>(interval_ms))
{}

void interval_timer::wait() {
	get()->wait();
}

} // namespace xtl
