#include <xtl/json.hpp>
#include <xtl/exception.hpp>
#include <inttypes.h>

namespace xtl {

inline bool constexpr is_whitespace(const char c) noexcept {
	return (c == ' ' || c == '\n' || c == '\r' || c == '\t');
}

inline bool constexpr is_digit(const char c) noexcept {
	return (c >= '0' && c <= '9');
}

inline char next(xtl::istream & stream) {
	char c;
	stream >> c;
	return c;
}

inline json::string_type to_string(json::number_type number) {
	json::string_type result;
	result.resize(snprintf(nullptr, 0, "%" PRId64, number));
	snprintf(&result[0], result.size() + 1, "%" PRId64, number);
	return result;
}

inline json::string_type to_string(bool boolean) {
	return (boolean) ? "true" : "false";
}

inline json::number_type to_number(const json::string_type & string) {
	return atoll(string.c_str());
}

inline json::string_type parse_string(xtl::istream & stream, char & c) {
	json::string_type string;
	for (;;) {
		stream >> c;
		if(c == '\\') {
			continue;
		} else {
			if (c == '"') {
				stream >> c;
				return string;
			} else {
				string += c;
			}
		}
	}
}

inline json::number_type parse_number(xtl::istream & stream, char & c) {
	json::string_type tmp;
	for (;;) {
		if (is_digit(c) || c == '-') {
			tmp += c;
			stream >> c;
		} else {
			return to_number(tmp);
		}
	}
}

inline bool parse_boolean(xtl::istream & stream, char & c) {
	if (c == 't') {
		const char r = next(stream);
		const char u = next(stream);
		const char e = next(stream);
		if (r == 'r' && u == 'u' && e == 'e') {
			stream >> c;
			return true;
		}
	} else {
		const char a = next(stream);
		const char l = next(stream);
		const char s = next(stream);
		const char e = next(stream);
		if (a == 'a' && l == 'l' && s == 's' && e == 'e') {
			stream >> c;
			return false;
		}
	}
	throw xtl::json_exception(__FILE__, __LINE__, "JSON error reading boolean");
}

inline void parse_null(xtl::istream & stream, char & c) {
	const char u = next(stream);
	const char l1 = next(stream);
	const char l2 = next(stream);
	if (u == 'u' && l1 == 'l' && l2 == 'l') {
		stream >> c;
		return;
	}
	throw xtl::json_exception(__FILE__, __LINE__, "JSON error reading null");
}

json::object_type parse_object(xtl::istream & stream, char & c);
json::array_type parse_array(xtl::istream & stream, char & c);

void parse_comma(xtl::istream & stream, char & c) {
	for (;;) {
		if (is_whitespace(c)) {
			stream >> c;
			continue;
		} else if (c == ',' || c == ']' || c == '}') {
			return;
		} else {
			throw xtl::json_exception(__FILE__, __LINE__, "JSON error reading comma");
		}
	}
}

void parse_colon(xtl::istream & stream, char & c) {
	for (;;) {
		if (c == ':') {
			stream >> c;
			return;
		} else if (is_whitespace(c)) {
			stream >> c;
		} else {
			throw xtl::json_exception(__FILE__, __LINE__, "JSON error reading colon");
		}
	}
}

bool parse_key(xtl::istream & stream, char & c, json::string_type & string) {
	for (;;) {
		if (c == '"') {
			string = parse_string(stream, c);
			return true;
		} else if (c == 't' || c == 'f') {
			string = to_string(parse_boolean(stream, c));
			return true;
		} else if (c == 'n') {
			parse_null(stream, c);
			string.clear();
			return true;
		} else if (is_digit(c) || c == '-') {
			string = to_string(parse_number(stream, c));
			return true;
		} else if (is_whitespace(c)) {
			stream >> c;
		} else if (c == '}') {
			return false;
		} else {
			throw xtl::json_exception(__FILE__, __LINE__, "JSON error reading key");
		}
	}
}

bool parse_value(xtl::istream & stream, char & c, json & js) {
	for (;;) {
		if (c == '{') {
			js = json(parse_object(stream, c));
			stream >> c;
			return true;
		} else if (c == '[') {
			js = json(parse_array(stream, c));
			stream >> c;
			return true;
		} else if (c == '"') {
			js = json(parse_string(stream, c));
			return true;
		} else if (c == 't' || c == 'f') {
			js = json(parse_boolean(stream, c));
			return true;
		} else if (c == 'n') {
			parse_null(stream, c);
			js = json(nullptr);
			return true;
		} else if (is_digit(c) || c == '-') {
			js = json(parse_number(stream, c));
			return true;
		} else if (is_whitespace(c)) {
			stream >> c;
		} else if (c == ']') {
			return false;
		} else {
			throw xtl::json_exception(__FILE__, __LINE__, "JSON error reading value");
		}
	}
}

json::object_type parse_object(xtl::istream & stream, char & c) {
	json::object_type object;
	json::string_type key;
	stream >> c;
	for (;;) {
		if (parse_key(stream, c, key)) {
			parse_colon(stream, c);
			if (!parse_value(stream, c, object[key])) {
				throw xtl::json_exception(__FILE__, __LINE__, "JSON error reading object");
			}
			parse_comma(stream, c);
			if (c == ',') {
				stream >> c;
				continue;
			} else if (c == '}') {
				return object;
			} else {
				throw xtl::json_exception(__FILE__, __LINE__, "JSON error reading object");
			}
		} else {
			return object;
		}
	}
}

json::array_type parse_array(xtl::istream & stream, char & c) {
	json::array_type array;
	json js;
	stream >> c;
	for (;;) {
		if (parse_value(stream, c, js)) {
			array.emplace_back(std::move(js));
			parse_comma(stream, c);
			if (c == ',') {
				stream >> c;
				continue;
			} else if (c == ']') {
				return array;
			} else {
				throw xtl::json_exception(__FILE__, __LINE__, "JSON error reading array");
			}
		} else {
			return array;
		}
	}
	return array;
}

json::json(xtl::istream & stream)
: m_type(type::unknown)
{
	for (;;) {
		char c = next(stream);
		if (c == '{') {
			new (&m_object) object_type(std::move(parse_object(stream, c)));
			m_type = type::object;
			return;
		} else if (c == '[') {
			new (&m_array) array_type(std::move(parse_array(stream, c)));
			m_type = type::array;
			return;
		} else if (!is_whitespace(c)) {
			throw xtl::json_exception(__FILE__, __LINE__, "JSON error reading root");
		}
	}
	throw xtl::json_exception(__FILE__, __LINE__, "JSON error reading root");
}

json::json(json && other) noexcept
: m_type(type::unknown)
{
	switch (other.m_type) {
		case type::array: { new (&m_array) array_type(std::move(other.m_array)); break; }
		case type::boolean: { m_boolean = other.m_boolean; break; }
		case type::unknown: { break; }
		case type::null: { break; }
		case type::number: { new (&m_number) number_type(std::move(other.m_number)); break; }
		case type::object: { new (&m_object) object_type(std::move(other.m_object)); break; }
		case type::string: { new (&m_string) string_type(std::move(other.m_string));  break; }
	}
	m_type = other.m_type;
}

json & json::operator=(json && other) noexcept {
	switch (m_type) {
		case type::array: { m_array.~array_type(); break; }
		case type::boolean: { break; }
		case type::unknown: { break; }
		case type::null: { break; }
		case type::number: { m_number.~number_type(); break; }
		case type::object: { m_object.~object_type(); break; }
		case type::string: { m_string.~string_type();  break; }
	}
	m_type = type::unknown;
	switch (other.m_type) {
		case type::array: { new (&m_array) array_type(std::move(other.m_array)); break; }
		case type::boolean: { m_boolean = other.m_boolean; break; }
		case type::unknown: { break; }
		case type::null: { break; }
		case type::number: { new (&m_number) number_type(std::move(other.m_number)); break; }
		case type::object: { new (&m_object) object_type(std::move(other.m_object)); break; }
		case type::string: { new (&m_string) string_type(std::move(other.m_string));  break; }
	}
	m_type = other.m_type;
	return *this;
}

json::~json() noexcept {
	switch (m_type) {
		case type::string: { m_string.~string_type(); break; }
		case type::object: { m_object.~object_type(); break; }
		case type::array: { m_array.~array_type(); break; }
		default: { break; }
	}
}

size_t json::size() const noexcept {
	switch (m_type) {
		case type::string: return m_string.length();
		case type::object: return m_object.size();
		case type::array: return m_array.size();
		default: return 0;
	}
}

const json & json::operator[](const size_t index) const {
	if (m_type == type::array) {
		if (index >= m_array.size()) {
			throw xtl::json_exception(__FILE__, __LINE__, "JSON array index out of bounds");
		}
		return m_array[index];
	}
	throw xtl::json_exception(__FILE__, __LINE__, "Invalid JSON type conversion requested");
}

const json & json::operator[](const string_type & key) const {
	if (m_type == type::object) {
		const auto it = m_object.find(key);
		if (it != m_object.end()) {
			return it->second;
		}
		throw xtl::json_exception(__FILE__, __LINE__, "Nonexistent property requested from JSON object");
	}
	throw xtl::json_exception(__FILE__, __LINE__, "Invalid JSON type conversion requested");
}

const json & json::operator[](const char * key) const {
	if (m_type == type::object) {
		const auto it = m_object.find(key);
		if (it != m_object.end()) {
			return it->second;
		}
		throw xtl::json_exception(__FILE__, __LINE__, "Nonexistent property requested from JSON object");
	}
	throw xtl::json_exception(__FILE__, __LINE__, "Invalid JSON type conversion requested");
}

json::string_type json::string() const {
	switch (m_type) {
		case type::string: return m_string;
		case type::number: return to_string(m_number);
		case type::boolean: return to_string(m_boolean);
		case type::null: return "null";
		default: throw xtl::json_exception(__FILE__, __LINE__, "Invalid JSON type conversion requested");
	}
}

int64_t json::number() const {
	switch (m_type) {
		case type::number: return m_number;
		case type::boolean: return m_boolean;
		case type::null: return 0;
		default: throw xtl::json_exception(__FILE__, __LINE__, "Invalid JSON type conversion requested");
	}
}

bool json::boolean() const {
	switch (m_type) {
		case type::number: return m_number;
		case type::boolean: return m_boolean;
		case type::null: return false;
		default: throw xtl::json_exception(__FILE__, __LINE__, "Invalid JSON type conversion requested");
	}
}

const json::object_type & json::object() const {
	if (m_type == type::object) {
		return m_object;
	}
	throw xtl::json_exception(__FILE__, __LINE__, "Invalid JSON type conversion requested");
}

const json::array_type & json::array() const {
	if (m_type == type::array) {
		return m_array;
	}
	throw xtl::json_exception(__FILE__, __LINE__, "Invalid JSON type conversion requested");
}

} // namespace xtl
