#include <xtl/mutex.hpp>

#include <xtl/exception.hpp>
#include <xtl/file_descriptor.hpp>
#include <xtl/scheduler.hpp>
#include <sys/eventfd.h>
#include <unistd.h>

#if defined(DEBUG_MUTEX) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) fprintf(stderr, "mutex[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) fprintf(stderr, "mutex[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xtl {

inline int make_mutex() {
	const int fd = ::eventfd(1, EFD_SEMAPHORE);
	if (fd < 0) {
		throw xtl::system_exception(__FILE__, __LINE__);
	}
	return fd;
}

class mutex_impl {
public:
	inline mutex_impl()
	: m_fd(make_mutex())
	{
		fdebug("Created new mutex %p", (void *) this);
		xtl::scheduler::monitor(m_fd);
	}

	inline ~mutex_impl() noexcept {
		fdebug("Destroying mutex %p", (void *) this);
		xtl::scheduler::release(m_fd);
	}

	inline void lock() {
		uint64_t value;
		for (;;) {
			const ssize_t result = ::read(m_fd.value(), &value, sizeof(value));
			if (result > 0) {
				fdebug("Acquired lock on mutex %p", (void *) this);
				return;
			} else if (result == 0) {
				throw xtl::end_of_stream(__FILE__, __LINE__);
			} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
				fdebug("Lock busy on mutex %p", (void *) this);
				xtl::scheduler::wait_for(m_fd);
			} else if (errno == EINTR) {
				continue;
			} else {
				throw xtl::system_exception(__FILE__, __LINE__);
			}
		}
	}

	inline void unlock() {
		const uint64_t value = 1;
		for (;;) {
			const ssize_t result = ::write(m_fd.value(), &value, sizeof(value));
			if (result > 0) {
				fdebug("Released lock on mutex %p", (void *) this);
				return;
			} else if (result == 0) {
				throw xtl::end_of_stream(__FILE__, __LINE__);
			} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
				fdebug("Overflow on mutex %p", (void *) this);
				xtl::scheduler::wait_for(m_fd);
			} else if (errno == EINTR) {
				continue;
			} else {
				throw xtl::system_exception(__FILE__, __LINE__);
			}
		}
	}

private:
	xtl::file_descriptor m_fd;
};

mutex::mutex()
: shared_ptr(std::make_shared<mutex_impl>())
{}

void mutex::lock() {
	get()->lock();
}

void mutex::unlock() {
	get()->unlock();
}

} // namespace xtl
