#include <xtl/os.hpp>

#include <xtl/file_descriptor.hpp>
#include <xtl/exception.hpp>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <netinet/in.h>
#include <net/if.h>
#include <unistd.h>
#include <cerrno>

namespace xtl {

namespace os {

struct netlink_address : sockaddr_nl {
	inline netlink_address(
		pid_t pid
	) noexcept {
		nl_family = AF_NETLINK;
		nl_pid = pid;
	}

	inline const struct sockaddr * cast() const noexcept {
		return reinterpret_cast<const struct sockaddr *>(this);
	}
};

struct generic_request : rtgenmsg {
	inline generic_request(
		uint8_t family
	) {
		rtgen_family = family;
	}
};

struct netlink_message_header : nlmsghdr {
	inline netlink_message_header(
		const generic_request & request,
		uint16_t type,
		uint16_t flags,
		uint32_t sequence_number,
		uint32_t pid
	) {
		nlmsg_len = NLMSG_LENGTH(sizeof(request));
		nlmsg_type = type;
		nlmsg_flags = flags;
		nlmsg_seq = sequence_number;
		nlmsg_pid = pid;
	}
};

struct iovector : iovec {
	template<typename T>
	inline iovector(
		T & value
	) {
		iov_base = &value;
		iov_len = sizeof(value);
	}
};

template<size_t N>
inline ssize_t iovecsize(
	const iovector (&vectors)[N]
) {
	size_t size = 0;
	for (size_t i = 0; i < N; ++i) {
		size += vectors[i].iov_len;
	}
	return size;
}

struct message_header : msghdr {
	template<size_t N>
	inline message_header(
		netlink_address & address,
		iovector (&vectors)[N]
	) {
		msg_name = &address;
		msg_namelen = sizeof(address);
		msg_iov = vectors;
		msg_iovlen = N;
		msg_control = nullptr;
		msg_controllen = 0;
		msg_flags = 0;
	}
};

inline int create_netlink_socket() {
	const int result = ::socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
	if (result < 0) {
		throw xtl::system_exception(__FILE__, __LINE__);
	}
	return result;
}

inline void parse_interface_message(
	const struct ifinfomsg * info_message,
	int message_length,
	std::vector<xtl::network_interface> & interfaces
) {
	typedef const xtl::mac_address * mac_address_ptr;
	typedef const char * ifname_ptr;
	if (!(info_message->ifi_flags & IFF_LOOPBACK)) {
		std::string ifname;
		for (
			const auto * attribute = IFLA_RTA(info_message);
			RTA_OK(attribute, message_length);
			attribute = RTA_NEXT(attribute, message_length)
		) {
			// ifname followed by ifaddress
			switch (attribute->rta_type) {
				case IFLA_IFNAME: {
					const auto attribute_length = RTA_PAYLOAD(attribute);
					ifname.assign(ifname_ptr(RTA_DATA(attribute)), attribute_length);
					break;
				}
				case IFLA_ADDRESS: {
					const auto attribute_length = RTA_PAYLOAD(attribute);
					if (attribute_length == sizeof(xtl::mac_address)) {
						interfaces.push_back({ ifname, *mac_address_ptr(RTA_DATA(attribute)) });
					}
					break;
				}
			}
		}
	}
}

inline bool parse_netlink_message(
	const struct nlmsghdr * message,
	std::vector<xtl::network_interface> & interfaces
) {
	switch (message->nlmsg_type) {
		case NLMSG_DONE: { return true; }
		case RTM_NEWLINK: {
			const auto * interface = static_cast<const struct ifinfomsg *>(NLMSG_DATA(message));
			const int message_length = NLMSG_PAYLOAD(message, sizeof(*interface));
			parse_interface_message(interface, message_length, interfaces);
			break;
		}
	}
	return false;
}

std::vector<xtl::network_interface> interfaces() {
	const pid_t pid = ::getpid();
	xtl::file_descriptor fd(create_netlink_socket());
	const netlink_address address(pid);
	if (::bind(fd.value(), address.cast(), sizeof(address)) != 0) {
		throw xtl::system_exception(__FILE__, __LINE__);
	}
	std::vector<xtl::network_interface> interfaces;
	netlink_address kernel(0);
	generic_request generic(AF_INET);
	netlink_message_header netlink_message(generic, RTM_GETLINK, NLM_F_REQUEST | NLM_F_DUMP, 1, pid);
	iovector output[2] = { iovector(netlink_message), iovector(generic) };
	const ssize_t output_size = iovecsize(output);
	message_header request(kernel, output);
	if (::sendmsg(fd.value(), &request, 0) != output_size) {
		throw xtl::system_exception(__FILE__, __LINE__);
	}
	char buffer[8192];
	iovector input[1] = { buffer };
	message_header response(kernel, input);
	for (bool done = false; !done;) {
		auto result = ::recvmsg(fd.value(), &response, 0);
		if (result < 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		} else if (result == 0) {
			done = true;
		} else {
			for (
				const auto * message = reinterpret_cast<const struct nlmsghdr *>(buffer);
				NLMSG_OK(message, result);
				message = NLMSG_NEXT(message, result)
			) {
				done = parse_netlink_message(message, interfaces);
			}
		}
	}
	return interfaces;
}

} // namespace os

} // namespace xtl
