#include <xtl/pipe.hpp>

#include <xtl/exception.hpp>
#include <xtl/file_descriptor.hpp>
#include <xtl/scheduler.hpp>
#include <utility>
#include <fcntl.h>
#include <unistd.h>

#if defined(DEBUG_PIPE) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) fprintf(stderr, "pipe[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) fprintf(stderr, "pipe[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xtl {

using fd_pair = std::pair<xtl::file_descriptor, xtl::file_descriptor>;

inline fd_pair make_pipe() {
	int fds[2];
	if (::pipe2(fds, O_CLOEXEC | O_NONBLOCK) != 0) {
		throw xtl::system_exception(__FILE__, __LINE__);
	}
	return fd_pair(fds[0], fds[1]);
}

class pipe_impl {
public:
	inline pipe_impl(fd_pair fds)
	: m_input(std::move(fds.second))
	, m_output(std::move(fds.first))
	{
		fdebug("Created new pipe_impl %p", (void *) this);
		scheduler::monitor(m_input);
		scheduler::monitor(m_output);
	}

	inline ~pipe_impl() noexcept {
		fdebug("Destroying pipe_impl %p", (void *) this);
		scheduler::release(m_input);
		scheduler::release(m_output);
	}

	inline size_t read(void * buffer, size_t amount) {
		for (;;) {
			const ssize_t result = ::read(m_output.value(), buffer, amount);
			if (result > 0) {
				return result;
			} else if (result == 0) {
				throw xtl::end_of_stream(__FILE__, __LINE__);
			} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
				xtl::scheduler::wait_for(m_output);
			} else if (errno == EINTR) {
				continue;
			} else {
				throw xtl::system_exception(__FILE__, __LINE__);
			}
		}
		return amount;
	}

	inline size_t write(const void * buffer, size_t amount) {
		for (;;) {
			const ssize_t result = ::write(m_input.value(), buffer, amount);
			if (result > 0) {
				return result;
			} else if (result == 0 || errno == EPIPE) {
				throw xtl::end_of_stream(__FILE__, __LINE__);
			} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
				xtl::scheduler::wait_for(m_input);
			} else if (errno == EINTR) {
				continue;
			} else {
				throw xtl::system_exception(__FILE__, __LINE__);
			}
		}
		return amount;
	}

	inline size_t flush() {
		return 0;
	}

private:
	xtl::file_descriptor m_input;
	xtl::file_descriptor m_output;
};

pipe::pipe()
: shared_ptr(std::make_shared<xtl::pipe_impl>(make_pipe()))
{}

size_t pipe::read(void * buffer, size_t amount) {
	return get()->read(buffer, amount);
}

size_t pipe::write(const void * buffer, size_t amount) {
	return get()->write(buffer, amount);
}

size_t pipe::flush() {
	return get()->flush();
}

} // namespace xtl
