#include <xtl/rsa.hpp>

#include <xtl/exception.hpp>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <cstring>

namespace xtl {

inline EVP_MD_CTX * make_rsa_verify() {
	EVP_MD_CTX * context = EVP_MD_CTX_create();
	if (context == nullptr) {
		throw xtl::rsa_exception(__FILE__, __LINE__, "EVP_MD_CTX_create error");
	}
	return context;
}

inline EVP_PKEY * make_public_key(const xtl::rsa_public_key & public_key) {
	const unsigned char * ptr = static_cast<const unsigned char *>(public_key.value());
	EVP_PKEY * key = d2i_PUBKEY(nullptr, &ptr, public_key.size());
	if (key == nullptr) {
		throw xtl::rsa_exception(__FILE__, __LINE__, "EVP_PKEY_new error");
	}
	return key;
}

class rsa_verify_impl {
public:
	inline rsa_verify_impl(const xtl::rsa_public_key & public_key)
	: m_context(make_rsa_verify())
	{
		EVP_PKEY * key = make_public_key(public_key);
		if (EVP_DigestVerifyInit(m_context, nullptr, EVP_sha256(), nullptr, key) == 0) {
			EVP_PKEY_free(key);
			throw xtl::rsa_exception(__FILE__, __LINE__, "EVP_DigestVerifyInit error");
		}
		EVP_PKEY_free(key);
	}

	inline ~rsa_verify_impl() {
		EVP_MD_CTX_destroy(m_context);
	}

	inline void append(const void * buffer, size_t amount) {
		if (EVP_DigestVerifyUpdate(m_context, buffer, amount) == 0) {
			throw xtl::rsa_exception(__FILE__, __LINE__, "EVP_DigestVerifyUpdate error");
		}
	}

	inline bool verify(const xtl::rsa_signature & signature) {
		unsigned char * ptr = static_cast<unsigned char *>(const_cast<void *>(signature.value())); // Braindead OpenSSL
		const int result = EVP_DigestVerifyFinal(m_context, ptr, signature.size());
		if (result > 0) {
			return true;
		} else if (result == 0) {
			return false;
		} else {
			throw xtl::rsa_exception(__FILE__, __LINE__, "EVP_DigestVerifyFinal error");
		}
	}

private:
	EVP_MD_CTX * m_context;
};

rsa_verify::rsa_verify(const xtl::rsa_public_key & public_key)
: shared_ptr(std::make_shared<rsa_verify_impl>(public_key)) {}

verify_method & rsa_verify::append(const void * buffer, size_t amount) {
	get()->append(buffer, amount);
	return *this;
}

bool rsa_verify::verify(const xtl::rsa_signature & signature) {
	return get()->verify(signature);

}

} // namespace xtl
