#include <xtl/scheduler.hpp>
#include <xtl/fiber.hpp>
#include <xtl/thread.hpp>

#include <xtl/exception.hpp>
#include <xtl/stack.hpp>
#include <unordered_map>
#include <unordered_set>
#include <map>
#include <deque>
#include <cstring>
#include <sys/epoll.h>
#include <ucontext.h>

#ifdef WITH_VALGRIND
#include <valgrind/valgrind.h>
#endif

#if defined(DEBUG_SCHEDULER) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) fprintf(stderr, "scheduler[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) fprintf(stderr, "scheduler[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xtl {

template<typename T>
union trampoline_arg {
public:
	inline constexpr trampoline_arg(T * pointer) noexcept : m_pointer(pointer) {}
	inline constexpr trampoline_arg(int a, int b) noexcept : m_ints{a, b} {}

	inline T * get() const noexcept {
		return m_pointer;
	}

	inline T * operator->() const noexcept {
		return m_pointer;
	}

	inline int operator[](size_t index) const noexcept {
		return m_ints[index];
	}

private:
	T * m_pointer;
	int m_ints[4];
};

class scheduler_impl;

struct execution_context {
	inline execution_context(xtl::scheduler_impl * scheduler)
	: m_scheduler(scheduler)
	, m_valgrind(-1)
	, m_blocked_by(-1)
	{}

	inline execution_context(
		xtl::function<void()> && function,
		size_t stack_size,
		xtl::scheduler_impl * scheduler,
		void (*trampoline) (void)
	)
	: m_function(std::move(function))
	, m_stack(stack_size)
	, m_scheduler(scheduler)
#ifdef WITH_VALGRIND
	, m_valgrind(VALGRIND_STACK_REGISTER(m_stack.get(), static_cast<char *>(m_stack.get()) + m_stack.size()))
#else
	, m_valgrind(-1)
#endif
	, m_blocked_by(-1)
	{
		if (::getcontext(&m_ucontext) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
		m_ucontext.uc_stack.ss_sp = m_stack.get();
		m_ucontext.uc_stack.ss_size = m_stack.size();
		m_ucontext.uc_link = nullptr;
		const xtl::trampoline_arg<xtl::execution_context> arg(this);
		::makecontext(&m_ucontext, trampoline, 2, arg[0], arg[1]);
		fdebug("(%p): Created context %p", (void *) scheduler, (void *) this);
	}

	inline ~execution_context() noexcept {
#ifdef WITH_VALGRIND
		if (m_valgrind >= 0) {
			VALGRIND_STACK_DEREGISTER(m_valgrind);
		}
#endif
		fdebug("(%p): Destroyed context %p", nullptr, (void *) this);
	}

	xtl::function<void()> m_function;
	ucontext_t m_ucontext;
	xtl::stack m_stack;
	xtl::scheduler_impl * m_scheduler;
	int m_valgrind;
	int m_blocked_by;
};

struct epoll_event : ::epoll_event {
	inline epoll_event(int fd) noexcept {
		memset(this, 0, sizeof(*this));
		events = EPOLLIN | EPOLLOUT | EPOLLERR | EPOLLHUP | EPOLLRDHUP | EPOLLET;
		data.fd = fd;
	}
};

inline int make_epoll() {
	const int fd = ::epoll_create1(EPOLL_CLOEXEC);
	if (fd < 0) {
		throw xtl::system_exception(__FILE__, __LINE__);
	}
	return fd;
}

class scheduler_impl {
public:
	inline scheduler_impl()
	: m_fd(make_epoll())
	, m_current_context(&m_thread_context)
	, m_scheduler_context(
		xtl::function<void()>([this] () { run(); }),
		16384,
		this,
		reinterpret_cast<void (*)()>(trampoline)
	)
	, m_thread_context(this)
	, m_active(true)
	{}

	inline ~scheduler_impl() noexcept {
		fdebug("Destroyed scheduler %p", (void *) this);
	}

	inline void monitor(const xtl::file_descriptor & descriptor) {
		const int fd = descriptor.value();
		if (m_blocked.find(fd) != m_blocked.end()) {
			return;
		}
		xtl::epoll_event event(fd);
		if (::epoll_ctl(m_fd.value(), EPOLL_CTL_ADD, fd, &event) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
		m_blocked.emplace(std::piecewise_construct, std::forward_as_tuple(fd), std::forward_as_tuple());
	}

	inline void release(const xtl::file_descriptor & descriptor) {
		const int fd = descriptor.value();
		const auto blocked = m_blocked.find(fd);
		if (blocked == m_blocked.end()) {
			return;
		}
		if (::epoll_ctl(m_fd.value(), EPOLL_CTL_DEL, fd, nullptr) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
		for (xtl::execution_context * context : blocked->second) {
			context->m_blocked_by = -1;
			m_tasklist.push_back(context);
		}
		m_blocked.erase(blocked);
	}

	inline void clear_timeout(xtl::execution_context *) {
	}

	inline void set_timeout(const xtl::file_descriptor &, size_t) {
	}

	inline void wait_for(const xtl::file_descriptor & descriptor) {
		const int fd = descriptor.value();
		fdebug("(%p): Context %p waiting for %d", (void *) this, (void *) m_current_context, fd);
		monitor(descriptor);
		m_current_context->m_blocked_by = fd;
		m_blocked[fd].emplace(m_current_context);
		switch_to(&m_scheduler_context);
	}

	inline void yield() {
		fdebug("(%p): Context %p yielding", (void *) this, (void *) m_current_context);
		m_tasklist.push_back(m_current_context);
		switch_to(&m_scheduler_context);
	}

	xtl::execution_context * create_context(
		xtl::function<void()> && function,
		size_t stack_size
	) {
		auto ptr = std::make_unique<xtl::execution_context>(
			std::move(function),
			stack_size,
			this,
			reinterpret_cast<void (*)()>(trampoline)
		);
		xtl::execution_context * context = m_contexts.emplace(
			std::piecewise_construct,
			std::forward_as_tuple(ptr.get()),
			std::forward_as_tuple(std::move(ptr))
		).first->first;
		m_tasklist.push_back(context);
		return context;
	}

	void destroy_context(xtl::execution_context * context) {
		fdebug("(%p): Marking %p as dead", (void *) this, (void *) context);
		if (context->m_blocked_by >= 0) {
			m_blocked[context->m_blocked_by].erase(context);
		}
		for (auto it = m_tasklist.begin(); it != m_tasklist.end();) {
			if (*it == context) {
				it = m_tasklist.erase(it);
			} else {
				++it;
			}
		}
		m_dead.push_back(context);
	}

private:
	xtl::file_descriptor m_fd;
	xtl::execution_context * m_current_context;
	xtl::execution_context m_scheduler_context;
	xtl::execution_context m_thread_context;
	std::unordered_map<xtl::execution_context *, std::unique_ptr<xtl::execution_context>> m_contexts;
	std::unordered_map<int, std::unordered_set<xtl::execution_context *>> m_blocked;
	std::deque<xtl::execution_context *> m_dead;
	std::deque<xtl::execution_context *> m_tasklist;
	bool m_active;

	static void trampoline(int a, int b) {
		const xtl::trampoline_arg<xtl::execution_context> arg(a, b);
		try {
			arg->m_function();
		} catch (const std::exception & e) {
			fprintf(stderr, "%s", e.what());
		}
		xtl::scheduler_impl * self = arg->m_scheduler;
		fdebug("(%p): Context %p terminated", (void *) self, (void *) arg.get());
		self->destroy_context(arg.get());
		self->switch_to(&self->m_scheduler_context);
		::abort(); // should never happen
	}

	inline void execute_tasklist() {
		while (!m_tasklist.empty()) {
			xtl::execution_context * context = m_tasklist.front();
			m_tasklist.pop_front();
			if (context) {
				switch_to(context);
			} else {
				return;
			}
		}
	}

	inline void reap_dead_contexts() {
		for (auto it = m_dead.begin(); it != m_dead.end(); ++it) {
			fdebug("(%p): Reaping %p", (void *) this, (void *) *it);
			m_contexts.erase(*it);
		}
		m_dead.clear();
	}

	void run() {
		for (;;) {
			m_tasklist.push_back(nullptr);
			execute_tasklist();
			reap_dead_contexts();
			wait_for_events();
		}
	}

	void switch_to(xtl::execution_context * context) {
		fdebug("(%p): Switching from %p to %p", (void *) this, (void *) m_current_context, (void *) context);
		if (::swapcontext(&std::exchange(m_current_context, context)->m_ucontext, &context->m_ucontext) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
	}

	inline size_t now() {
		struct timespec res;
		if (clock_gettime(CLOCK_MONOTONIC, &res) != 0) {
			throw xtl::system_exception(__FILE__, __LINE__);
		}
		return (uint64_t(res.tv_sec) * 1000) + (res.tv_nsec / 1000000);
	}

	void wait_for_events() {
		const size_t max_events = 16;
		struct ::epoll_event events[max_events];
		for (;;) {
			int timeout = -1;
			const int result = ::epoll_wait(m_fd.value(), events, max_events, timeout);
			if (result > 0) {
				for (int i = 0; i < result; ++i) {
					const auto fd = events[i].data.fd;
					const auto blocked = m_blocked.find(fd);
					if (blocked != m_blocked.end()) {
						for (xtl::execution_context * context : blocked->second) {
							context->m_blocked_by  = -1;
							m_tasklist.push_back(context);
							fdebug("(%p): Unblocked %p", (void *) this, (void *) context);
						}
					}
				}
			} else if (result == 0) {
			} else if (errno == EINTR) {
				continue;
			} else {
				throw xtl::system_exception(__FILE__, __LINE__);
			}
			return;
		}
	}
};

pthread_key_t g_sched_key;
pthread_once_t g_sched_once = PTHREAD_ONCE_INIT;

scheduler_impl & current_scheduler() {
	pthread_once(&g_sched_once, [] () {
		pthread_key_create(&g_sched_key, [] (void * ptr) {
			delete static_cast<scheduler_impl *>(ptr);
		});
	});
	scheduler_impl * scheduler = static_cast<scheduler_impl *>(pthread_getspecific(g_sched_key));
	if (scheduler == nullptr) {
		scheduler = new scheduler_impl();
		pthread_setspecific(g_sched_key, scheduler);
	}
	return *scheduler;
}

void scheduler::monitor(const xtl::file_descriptor & fd) {
	current_scheduler().monitor(fd);
}

void scheduler::release(const xtl::file_descriptor & fd) {
	current_scheduler().release(fd);
}

void scheduler::set_timeout(const xtl::file_descriptor & fd, size_t timeout_ms) {
	current_scheduler().set_timeout(fd, timeout_ms);
}

void scheduler::wait_for(const xtl::file_descriptor & fd) {
	current_scheduler().wait_for(fd);
}

void yield() {
	current_scheduler().yield();
}

void * scheduler::create_fiber(xtl::function<void()> && function, size_t stack_size) {
	return current_scheduler().create_context(std::move(function), stack_size);
}

void scheduler::destroy_fiber(void * fiber) {
	current_scheduler().destroy_context(static_cast<xtl::execution_context *>(fiber));
}

} // namespace xtl
