#include <xtl/sha256.hpp>

#include <openssl/evp.h>
#include <xtl/exception.hpp>
#include <cstring>

namespace xtl {

bool sha256_digest::operator==(const xtl::sha256_digest & other) const noexcept {
	return (memcmp(m_value, other.m_value, sizeof(m_value)) == 0);
}

bool sha256_digest::operator!=(const xtl::sha256_digest & other) const noexcept {
	return (memcmp(m_value, other.m_value, sizeof(m_value)) != 0);
}

xtl::ostream & operator<<(xtl::ostream & stream, const xtl::sha256_digest & digest) {
	xtl::write(stream, digest.m_value, sizeof(digest.m_value));
	return stream;
}

xtl::istream & operator>>(xtl::istream & stream, xtl::sha256_digest & digest) {
	xtl::read(stream, digest.m_value, sizeof(digest.m_value));
	return stream;
}

inline EVP_MD_CTX * make_sha256() {
	EVP_MD_CTX * context = EVP_MD_CTX_create();
	if (context == nullptr) {
		throw xtl::sha256_exception(__FILE__, __LINE__, "EVP_MD_CTX_create error");
	}
	return context;
}

class sha256_impl {
public:
	inline sha256_impl()
	: m_context(make_sha256())
	{
		if (EVP_DigestInit_ex(m_context, EVP_sha256(), nullptr) == 0) {
			throw xtl::sha256_exception(__FILE__, __LINE__, "EVP_DigestInit_ex error");
		}
	}

	inline ~sha256_impl() noexcept {
		EVP_MD_CTX_destroy(m_context);
	}

	inline void append(const void * buffer, size_t amount) {
		if (EVP_DigestUpdate(m_context, buffer, amount) == 0) {
			throw xtl::sha256_exception(__FILE__, __LINE__, "EVP_DigestUpdate error");
		}
	}

	inline xtl::sha256_digest digest() const {
		xtl::sha256_digest digest;
		unsigned int len = digest.size();
		if (EVP_DigestFinal_ex(m_context, static_cast<uint8_t *>(digest.value()), &len) == 0) {
			throw xtl::sha256_exception(__FILE__, __LINE__, "EVP_DigestFinal_ex error");
		}
		return digest;
	}

private:
	EVP_MD_CTX * m_context;
};

sha256::sha256()
: shared_ptr(std::make_shared<sha256_impl>())
{}

xtl::hash_method & sha256::append(const void * buffer, size_t amount) {
	get()->append(buffer, amount);
	return *this;
}

xtl::sha256_digest sha256::digest() const {
	return get()->digest();
}

} // namespace xtl
