#include <xtl/unix_acceptor.hpp>
#include <xtl/unix_socket.hpp>

#include <xtl/exception.hpp>
#include <xtl/file_descriptor.hpp>
#include <xtl/scheduler.hpp>

#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <unistd.h>

#include <cstring>
#include <cerrno>
#include <memory>

#if defined(DEBUG_UNIX_SOCKET) || defined(DEBUG_ALL)
#define fdebug(fmt, ...) fprintf(stderr, "unix_socket[%d]: " fmt "\n", __LINE__, __VA_ARGS__);
#define debug(msg) fprintf(stderr, "unix_socket[%d]: " msg "\n", __LINE__);
#else
#define fdebug(fmt, ...)
#define debug(msg)
#endif

namespace xtl {

class unix_address : private sockaddr_un {
public:
	inline unix_address(const std::string & filename) noexcept {
		sun_family = AF_UNIX;
		::strncpy(sun_path, filename.c_str(), sizeof(sun_path) - 1);
	}

	inline const sockaddr * cast() const noexcept {
		return reinterpret_cast<const sockaddr *>(this);
	}

	inline static size_t size() noexcept {
		return sizeof(sockaddr_un);
	}
};

inline int make_unix_socket() {
	const int fd = ::socket(AF_UNIX, SOCK_STREAM | SOCK_NONBLOCK | SOCK_CLOEXEC, 0);
	if (fd < 0) {
		throw xtl::unix_socket_exception(__FILE__, __LINE__);
	}
	return fd;
}

class unix_socket_impl {
public:
	inline unix_socket_impl(xtl::file_descriptor & fd) noexcept
	: m_fd(std::move(fd))
	{
		fdebug("Created new unix_socket %p", (void *) this);
		xtl::scheduler::monitor(m_fd);
	}

	inline unix_socket_impl(const std::string & filename)
	: m_fd(make_unix_socket())
	{
		fdebug("Created new unix_socket %p", (void *) this);
		if (::connect(m_fd.value(), xtl::unix_address(filename).cast(), xtl::unix_address::size()) != 0) {
			throw xtl::unix_socket_exception(__FILE__, __LINE__);
		}
		xtl::scheduler::monitor(m_fd);
	}

	inline ~unix_socket_impl() noexcept {
		xtl::scheduler::release(m_fd);
		fdebug("Destroyed unix_socket %p", (void *) this);
	}

	inline size_t read(void * buffer, size_t amount) {
		for (;;) {
			const ssize_t result = ::recv(m_fd.value(), buffer, amount, 0);
			if (result > 0) {
				return result;
			} else if (result == 0) {
				throw xtl::end_of_stream(__FILE__, __LINE__);
			} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
				xtl::scheduler::wait_for(m_fd);
			} else if (errno == EINTR) {
				continue;
			} else {
				throw xtl::unix_socket_exception(__FILE__, __LINE__);
			}
		}
	}

	inline size_t write(const void * buffer, size_t amount) {
		for (;;) {
			const ssize_t result = ::send(m_fd.value(), buffer, amount, MSG_NOSIGNAL);
			if (result > 0) {
				return result;
			} else if (result == 0 || errno == EPIPE) {
				throw xtl::end_of_stream(__FILE__, __LINE__);
			} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
				xtl::scheduler::wait_for(m_fd);
			} else if (errno == EINTR) {
				continue;
			} else {
				throw xtl::unix_socket_exception(__FILE__, __LINE__);
			}
		}
	}

	inline size_t flush() {
		return 0;
	}

	inline void set_timeout(size_t timeout_ms) {
		xtl::scheduler::set_timeout(m_fd, timeout_ms);
	}

private:
	xtl::file_descriptor m_fd;
};

unix_socket::unix_socket(const std::string & filename)
: shared_ptr(std::make_shared<xtl::unix_socket_impl>(filename))
{}

unix_socket::unix_socket(xtl::file_descriptor fd)
: shared_ptr(std::make_shared<xtl::unix_socket_impl>(fd))
{}

void unix_socket::set_timeout(size_t timeout_ms) {
	get()->set_timeout(timeout_ms);
}

size_t unix_socket::read(void * buffer, size_t amount) {
	return get()->read(buffer, amount);
}

size_t unix_socket::write(const void * buffer, size_t amount) {
	return get()->write(buffer, amount);
}

size_t unix_socket::flush() {
	return get()->flush();
}

class unix_acceptor_impl {
public:
	inline unix_acceptor_impl(const std::string & filename)
	: m_fd(make_unix_socket())
	{
		fdebug("Created new unix_acceptor %p", (void *) this);
		if (::bind(m_fd.value(), xtl::unix_address(filename).cast(), xtl::unix_address::size()) != 0) {
			throw xtl::unix_socket_exception(__FILE__, __LINE__);
		}
		if (chmod(filename.c_str(), 0777) != 0) {
			throw xtl::unix_socket_exception(__FILE__, __LINE__);
		}
		if (::listen(m_fd.value(), SOMAXCONN) != 0) {
			throw xtl::unix_socket_exception(__FILE__, __LINE__);
		}
		scheduler::monitor(m_fd);
	}

	inline ~unix_acceptor_impl() noexcept {
		scheduler::release(m_fd);
		fdebug("Destroyed unix_acceptor %p", (void *) this);
	}

	inline xtl::unix_socket accept() {
		for (;;) {
			const int result = ::accept4(m_fd.value(), nullptr, nullptr, SOCK_NONBLOCK | SOCK_CLOEXEC);
			if (result >= 0) {
				return xtl::unix_socket(xtl::file_descriptor(result));
			} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
				xtl::scheduler::wait_for(m_fd);
			} else if (errno == EINTR) {
				continue;
			} else {
				throw xtl::unix_socket_exception(__FILE__, __LINE__);
			}
		}
	}

private:
	xtl::file_descriptor m_fd;
};


unix_acceptor::unix_acceptor(const std::string & filename)
: shared_ptr(std::make_shared<xtl::unix_acceptor_impl>(filename))
{}

xtl::unix_socket unix_acceptor::accept() {
	return get()->accept();
}

} // namespace xtl
